"""
An expression language for Raco: functions, booleans, aggregates, etc.

Most non-trivial operators and functions are in separate files in this package.
"""

from raco.utility import Printable
from raco import types

from abc import ABCMeta, abstractmethod

import logging
LOG = logging.getLogger(__name__)


class TypeSafetyViolation(Exception):
    pass


def check_type(_type, allowed_types):
    if _type not in allowed_types:
        raise TypeSafetyViolation("Type %s not among %s" % (
            _type, allowed_types))


def check_is_numeric(_type):
    check_type(_type, types.NUMERIC_TYPES)


class Expression(Printable):
    __metaclass__ = ABCMeta
    literals = None

    @abstractmethod
    def typeof(self, scheme, state_scheme):
        """Returns a string describing the expression's return type.

        :param scheme: The schema of the relation corresponding to this
        expression
        :param state_scheme: The schema of the state corresponding to this
        expression; this is None except for the StatefulApply operator.
        :return: A string from among types.type_names.
        """

    @classmethod
    def opstr(cls):
        """Return this operator as it would be printed in a string."""
        if not cls.literals:
            return cls.opname()
        return cls.literals[0]

    @abstractmethod
    def evaluate(self, _tuple, scheme, state=None):
        """Evaluate an expression in the context of a given tuple and schema.

        This is used for unit tests written against the fake database.
        """

    @abstractmethod
    def get_children(self):
        """Return a list of child expressions."""

    def __copy__(self):
        raise RuntimeError("Shallow copy not supported for expressions")

    def postorder(self, f):
        """Apply a function to each node in an expression tree.

        The function argument should return a scalar.  The return value
        of postorder is an iterator over these scalars.
        """
        yield f(self)

    def walk(self):
        """Visit the nodes in an expression tree.

        The return value is an iterator over the tree nodes.  The order is
        unspecified.
        """
        yield self

    @abstractmethod
    def apply(self, f):
        """Replace children with the result of a function"""

    def add_offset_by_terms(self, termsToOffset):
        """Add a constant offset to every positional reference in this tree
        using the map of terms to offset.

        This function assumes that every AttributeRef has been labeled with the
        term that it refers to"""

        for ex in self.walk():
            if isinstance(ex, UnnamedAttributeRef):
                offset = termsToOffset[ex.myTerm]
                LOG.debug("adding offset %s to %s", offset, ex)
                ex.position += offset

    def accept(self, visitor):
        """
        Default visitor accept method. Probably does
        not need to be overridden by leaves, but certainly
        by inner tree nodes.
        """
        visitor.visit(self)


class ZeroaryOperator(Expression):

    def __init__(self):
        pass

    def __eq__(self, other):
        return self.__class__ == other.__class__

    def __hash__(self):
        return hash(self.__class__)

    def __str__(self):
        return "%s" % self.opstr()

    def __repr__(self):
        return "{op}()".format(op=self.opname())

    def apply(self, f):
        pass

    def walk(self):
        yield self

    def get_children(self):
        return []


class UnaryOperator(Expression):

    def __init__(self, input):
        self.input = input

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.input == other.input

    def __hash__(self):
        return hash(self.__class__) + hash(self.input)

    def __str__(self):
        return "%s%s" % (self.opstr(), self.input)

    def __repr__(self):
        return "{op}({inp!r})".format(op=self.opname(), inp=self.input)

    def postorder(self, f):
        for x in self.input.postorder(f):
            yield x
        yield f(self)

    def apply(self, f):
        self.input = f(self.input)

    def walk(self):
        yield self
        for ex in self.input.walk():
            yield ex

    def accept(self, visitor):
        """For post order stateful visitors"""
        self.input.accept(visitor)
        visitor.visit(self)

    def typeof(self, scheme, state_scheme):
        return self.input.typeof(scheme, state_scheme)

    def get_children(self):
        return [self.input]


class BinaryOperator(Expression):

    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __eq__(self, other):
        return self.__class__ == other.__class__ and \
            self.left == other.left and self.right == other.right

    def __hash__(self):
        return hash(self.__class__) + hash(self.left) + hash(self.right)

    def __str__(self):
        return "(%s %s %s)" % (self.left, self.opstr(), self.right)

    def __repr__(self):
        return "{op}({l!r}, {r!r})".format(op=self.opname(), l=self.left,
                                           r=self.right)

    def postorder(self, f):
        for x in self.left.postorder(f):
            yield x
        for x in self.right.postorder(f):
            yield x
        yield f(self)

    def apply(self, f):
        self.left = f(self.left)
        self.right = f(self.right)

    def walk(self):
        yield self
        for ex in self.left.walk():
            yield ex
        for ex in self.right.walk():
            yield ex

    def accept(self, visitor):
        """
        For post-order stateful visitors
        """
        self.left.accept(visitor)
        self.right.accept(visitor)
        visitor.visit(self)

    def typeof(self, scheme, state_scheme):
        lt = types.LONG_TYPE
        ft = types.DOUBLE_TYPE
        type_map = {(lt, lt): lt, (lt, ft): ft, (ft, lt): ft, (ft, ft): ft}

        left_type = self.left.typeof(scheme, state_scheme)
        right_type = self.right.typeof(scheme, state_scheme)

        if (left_type, right_type) in type_map:
            return type_map[(left_type, right_type)]
        else:
            raise TypeSafetyViolation("Can't combine %s, %s for %s" % (
                left_type, right_type, self.__class__))

    def get_children(self):
        return [self.left, self.right]


class NaryOperator(Expression):

    def __init__(self, operands):
        self.operands = operands

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        return self.operands == other.operands

    def __hash__(self):
        return hash(self.__class__) + hash(self.operands)

    def __str__(self):
        return "(%s %s)" % \
            (self.opstr(), " ".join([str(i) for i in self.operands]))

    def postorder(self, f):
        for op in self.operands:
            for x in op.postorder(f):
                yield x
        yield f(self)

    def apply(self, f):
        self.operands = [f(op) for op in self.operands]

    def walk(self):
        yield self
        for op in self.operands:
            for ex in op.walk():
                yield ex

    def accept(self, visitor):
        """
        For post-order stateful visitors
        """
        for op in self.operands:
            op.accept(visitor)
        visitor.visit(self)

    def get_children(self):
        return self.operands


class Literal(ZeroaryOperator):

    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.value == other.value

    def __hash__(self):
        return hash(self.__class__) + hash(self.value)

    @staticmethod
    def vars():
        return []

    def __str__(self):
        return str(self.value)

    def typeof(self, scheme, state_scheme):
        return types.python_type_map[type(self.value)]

    def evaluate(self, _tuple, scheme, state=None):
        return self.value

    def apply(self, f):
        pass

    def get_children(self):
        return []

    def __repr__(self):
        return "{op}({val!r})".format(op=self.opname(), val=self.value)


class StringLiteral(Literal):
    def __str__(self):
        return '"{val}"'.format(val=self.value)


class NumericLiteral(Literal):
    pass


class BooleanLiteral(Literal):
    pass


class AttributeRef(Expression):

    def evaluate(self, _tuple, scheme, state=None):
        return _tuple[self.get_position(
            scheme, state.scheme if state else None)]

    @abstractmethod
    def get_position(self, scheme, state_scheme=None):
        """Return the position of the referenced attribute in the given
        scheme"""

    def apply(self, f):
        pass

    def walk(self):
        yield self

    def typeof(self, scheme, state_scheme):
        return scheme.getType(self.get_position(scheme, state_scheme))

    def get_children(self):
        return []


class NamedAttributeRef(AttributeRef):

    def __init__(self, attributename):
        self.name = attributename

    def __str__(self):
        return "%s" % (self.name)

    def __repr__(self):
        return "{op}({att!r})".format(op=self.opname(), att=self.name)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return type(self) == type(other) and self.name == other.name

    def get_position(self, scheme, state_scheme=None):
        return scheme.getPosition(self.name)


class UnnamedAttributeRef(AttributeRef):

    def __init__(self, position, debug_info=None):
        self.debug_info = debug_info
        self.position = position

    def __str__(self):
        if not self.debug_info:
            return "${pos}".format(pos=self.position)
        return "{dbg}".format(dbg=self.debug_info)

    def __repr__(self):
        return "{op}({pos!r}, {dbg!r})".format(op=self.opname(),
                                               pos=self.position,
                                               dbg=self.debug_info)

    def __eq__(self, other):
        return (other.__class__ == self.__class__
                and other.position == self.position)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.position)

    def __cmp__(self, other):
        assert other.__class__ == self.__class__
        return cmp(self.position, other.position)

    def get_position(self, scheme, state_scheme=None):
        return self.position


class StateRef(Expression):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_position(self, scheme, state_scheme):
        """Return the position of the referenced attribute in the given
        scheme"""

    def apply(self, f):
        pass

    def get_children(self):
        return []


class UnnamedStateAttributeRef(StateRef):

    def __init__(self, position):
        self.position = position

    def __str__(self):
        return "$%s" % (self.position)

    def __repr__(self):
        return "{op}({pos!r})".format(op=self.opname(), pos=self.position)

    def get_position(self, scheme, state_scheme):
        return self.position

    def evaluate(self, _tuple, scheme, state):
        return state.values[self.position]

    def typeof(self, scheme, state_scheme):
        assert state_scheme is not None
        return state_scheme.getType(self.position)


class NamedStateAttributeRef(StateRef):

    def __init__(self, attributename):
        self.name = attributename

    def __str__(self):
        return "%s" % (self.name)

    def __repr__(self):
        return "{op}({att!r})".format(op=self.opname(), att=self.name)

    def evaluate(self, _tuple, scheme, state):
        return state.values[self.get_position(scheme, state.scheme)]

    def get_position(self, scheme, state_scheme):
        return state_scheme.getPosition(self.name)

    def typeof(self, scheme, state_scheme):
        assert state_scheme is not None
        return state_scheme.getType(self.get_position(scheme, state_scheme))


class UDF(NaryOperator):
    pass


class PLUS(BinaryOperator):
    literals = ["+"]

    def evaluate(self, _tuple, scheme, state=None):
        return (self.left.evaluate(_tuple, scheme, state) +
                self.right.evaluate(_tuple, scheme, state))


class MINUS(BinaryOperator):
    literals = ["-"]

    def evaluate(self, _tuple, scheme, state=None):
        return (self.left.evaluate(_tuple, scheme, state) -
                self.right.evaluate(_tuple, scheme, state))


class DIVIDE(BinaryOperator):
    literals = ["/"]

    def evaluate(self, _tuple, scheme, state=None):
        return (float(self.left.evaluate(_tuple, scheme, state)) /
                self.right.evaluate(_tuple, scheme, state))

    def typeof(self, scheme, state_scheme):
        check_is_numeric(self.left.typeof(scheme, state_scheme))
        check_is_numeric(self.right.typeof(scheme, state_scheme))
        return types.DOUBLE_TYPE


class IDIVIDE(BinaryOperator):
    literals = ["//"]

    def evaluate(self, _tuple, scheme, state=None):
        return int(self.left.evaluate(_tuple, scheme, state) /
                   self.right.evaluate(_tuple, scheme, state))

    def typeof(self, scheme, state_scheme):
        check_is_numeric(self.left.typeof(scheme, state_scheme))
        check_is_numeric(self.right.typeof(scheme, state_scheme))
        return types.LONG_TYPE


class TIMES(BinaryOperator):
    literals = ["*"]

    def evaluate(self, _tuple, scheme, state=None):
        return (self.left.evaluate(_tuple, scheme, state) *
                self.right.evaluate(_tuple, scheme, state))


class CAST(UnaryOperator):
    def __init__(self, _type, input):
        """Initialize a cast operator.

        @param _type: A string denoting a type; must be from
        types.ALL_TYPES
        """
        assert _type in types.ALL_TYPES
        self._type = _type
        UnaryOperator.__init__(self, input)

    def __str__(self):
        return "%s(%s, %s)" % (self.opstr(), self._type, self.input)

    def __repr__(self):
        return "{op}({t!r}, {inp!r})".format(op=self.opname(), t=self._type,
                                             inp=self.input)

    def evaluate(self, _tuple, scheme, state=None):
        pytype = types.reverse_python_type_map[self.typeof(None, None)]
        return pytype(self.input.evaluate(_tuple, scheme, state))

    def typeof(self, scheme, state_scheme):
        # Note the lack of type-checking here; I didn't want to codify a
        # particular set of casting rules.
        return types.map_type(self._type)


class NEG(UnaryOperator):
    literals = ["-"]

    def evaluate(self, _tuple, scheme, state=None):
        return -1 * self.input.evaluate(_tuple, scheme, state)

    def typeof(self, scheme, state_scheme):
        input_type = self.input.typeof(scheme, state_scheme)
        check_is_numeric(input_type)
        return input_type


class Unbox(ZeroaryOperator):

    def __init__(self, relational_expression, field):
        """Initialize an unbox expression.

        relational_expression is a Myrial AST that evaluates to a relation.

        field is an optional column name/index within the relation.  If None,
        the system uses index 0.
        """
        self.relational_expression = relational_expression
        self.field = field

    def evaluate(self, _tuple, scheme, state=None):
        """Raise an error on attempted evaluation.

        Unbox expressions are not "evaluated" in the usual sense.  Rather, they
        are replaced with raw attribute references at evaluation time.
        """
        raise NotImplementedError()

    def typeof(self, scheme, state_scheme):
        raise NotImplementedError()  # See above comment

    def __repr__(self):
        return "{op}({re!r}, {f!r})".format(
            op=self.opname(), re=self.relational_expression, f=self.field)

    def __str__(self):
        return "{op}({re}.{f})".format(
            op=self.opname(), re=self.relational_expression,
            f=self.field or "$0")


class Case(Expression):

    def __init__(self, when_tuples, else_expr):
        """Initialize a Case expression.

        :param when_tuples: A list of tuples of the form
        (test_expr, result_expr)
        :type when_tuples: List of tuples of (Expression, Expression)
        :param else_expr: An expression to evaluate if no when clause is
        satisfied.
        :type else_expr: Expression
        """
        self.when_tuples = when_tuples
        self.else_expr = else_expr

    def evaluate(self, _tuple, scheme, state=None):
        for test_expr, result_expr in self.when_tuples:
            if test_expr.evaluate(_tuple, scheme, state):
                return result_expr.evaluate(_tuple, scheme, state)
        return self.else_expr.evaluate(_tuple, scheme, state)

    def postorder(self, f):
        for test_expr, result_expr in self.when_tuples:
            for x in test_expr.postorder(f):
                yield x
            for x in result_expr.postorder(f):
                yield x
        for x in self.else_expr.postorder(f):
            yield x
        yield f(self)

    def apply(self, f):
        self.when_tuples = [(f(test), f(result)) for test, result
                            in self.when_tuples]
        self.else_expr = f(self.else_expr)

    def walk(self):
        yield self
        for test_expr, result_expr in self.when_tuples:
            for x in test_expr.walk():
                yield x
            for x in result_expr.walk():
                yield x
        for x in self.else_expr.walk():
            yield x

    def get_children(self):
        test_exprs, result_exprs = zip(*self.when_tuples)
        return list(test_exprs) + list(result_exprs) + [self.else_expr]

    def to_binary(self):
        """Convert n-ary case statements to a binary case statement."""
        assert len(self.when_tuples) > 0
        if len(self.when_tuples) == 1:
            return self
        else:
            new_when_tuples = [self.when_tuples[0]]
            new_else = Case(self.when_tuples[1:], self.else_expr)
            return Case(new_when_tuples, new_else)

    def __str__(self):
        when_strs = ['WHEN %s THEN %s' % (test, result)
                     for test, result in self.when_tuples]
        return "CASE(%s ELSE %s)" % (' '.join(when_strs), self.else_expr)

    def __repr__(self):
        return "{op}({wt!r}, {els!r})".format(op=self.opname(),
                                              wt=self.when_tuples,
                                              els=self.else_expr)

    def typeof(self, scheme, state_scheme):
        all_exprs = [res_expr for test_expr, res_expr in self.when_tuples]
        all_exprs.append(self.else_expr)
        types = [ex.typeof(scheme, state_scheme) for ex in all_exprs]
        if len(set(types)) != 1:
            raise TypeSafetyViolation(
                "Case expresssions must resolve to a single type")
        return types[0]


import abc


class ExpressionVisitor(object):
    # TODO: make this more complete for kinds of expressions

    __metaclass__ = abc.ABCMeta

    def visit(self, expr):
        # use expr to dispatch to appropriate visit_* method
        typename = type(expr).__name__
        dispatchTo = getattr(self, "visit_%s" % (typename,))
        return dispatchTo(expr)

    @abc.abstractmethod
    def visit_NOT(self, unaryExpr):
        return

    @abc.abstractmethod
    def visit_AND(self, binaryExpr):
        return

    @abc.abstractmethod
    def visit_OR(self, binaryExpr):
        return

    @abc.abstractmethod
    def visit_EQ(self, binaryExpr):
        return

    @abc.abstractmethod
    def visit_NEQ(self, binaryExpr):
        return

    @abc.abstractmethod
    def visit_GT(self, binaryExpr):
        return

    @abc.abstractmethod
    def visit_LT(self, binaryExpr):
        return

    @abc.abstractmethod
    def visit_GTEQ(self, binaryExpr):
        return

    @abc.abstractmethod
    def visit_LTEQ(self, binaryExpr):
        return

    @abc.abstractmethod
    def visit_NamedAttributeRef(self, named):
        return

    @abc.abstractmethod
    def visit_UnnamedAttributeRef(self, unnamed):
        return

    @abc.abstractmethod
    def visit_StringLiteral(self, stringLiteral):
        return

    @abc.abstractmethod
    def visit_NumericLiteral(self, numericLiteral):
        return

    @abc.abstractmethod
    def visit_DIVIDE(self, binaryExpr):
        return

    @abc.abstractmethod
    def visit_PLUS(self, binaryExpr):
        return

    @abstractmethod
    def visit_MINUS(self, binaryExpr):
        return

    @abstractmethod
    def visit_IDIVIDE(self, binaryExpr):
        return

    @abstractmethod
    def visit_TIMES(self, binaryExpr):
        return

    @abstractmethod
    def visit_NEG(self, unaryExpr):
        return


class SimpleExpressionVisitor(ExpressionVisitor):
    @abstractmethod
    def visit_unary(self, unaryexpr):
        pass

    @abstractmethod
    def visit_binary(self, binaryexpr):
        pass

    @abstractmethod
    def visit_zeroary(self, zeroaryexpr):
        pass

    def visit_literal(self, literalexpr):
        pass

    @abstractmethod
    def visit_nary(self, naryexpr):
        pass

    def visit_attr(self, attr):
        pass

    def visit_NOT(self, unaryExpr):
        self.visit_unary(unaryExpr)

    def visit_AND(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_OR(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_EQ(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_NEQ(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_GT(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_LT(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_GTEQ(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_LTEQ(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_NamedAttributeRef(self, named):
        self.visit_attr(named)

    def visit_UnnamedAttributeRef(self, unnamed):
        self.visit_attr(unnamed)

    def visit_StringLiteral(self, stringLiteral):
        self.visit_literal(stringLiteral)

    def visit_NumericLiteral(self, numericLiteral):
        self.visit_literal(numericLiteral)

    def visit_DIVIDE(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_PLUS(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_MINUS(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_IDIVIDE(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_TIMES(self, binaryExpr):
        self.visit_binary(binaryExpr)

    def visit_NEG(self, unaryExpr):
        self.visit_unary(unaryExpr)



import unittest

from raco.utility import real_str
from collections import OrderedDict


class TestUtility(unittest.TestCase):
    def test_real_str_int(self):
        self.assertEqual(real_str(1), str(1))
        self.assertEqual(real_str(1, skip_out=True), str(1))

    def test_real_str_string(self):
        self.assertEqual(real_str("abc"), str("abc"))
        self.assertEqual(real_str("abc", skip_out=True), str("abc"))

    def test_real_str_list(self):
        self.assertEqual(real_str([1, 2]), "[1,2]")
        self.assertEqual(real_str([1, 2], skip_out=True), "1,2")

    def test_real_str_dict(self):
        d = OrderedDict([(1, 2), (3, 4)])
        self.assertEqual(real_str(d), "{1:2,3:4}")
        self.assertEqual(real_str(d, skip_out=True), "1:2,3:4")

    def test_real_str_set(self):
        # Python has no built-in ordered set, so we do not know the item order
        self.assertIn(real_str({1, 2}), ["{1,2}", "{2,1}"])
        self.assertIn(real_str({1, 2}, skip_out=True), ["1,2", "2,1"])

    def test_real_str_nested_collections(self):
        self.assertEqual(real_str([[1, 2], {3: 4}, []]),
                         "[[1,2],{3:4},[]]")
        self.assertEqual(real_str([[1, 2], {3: 4}, []], skip_out=True),
                         "[1,2],{3:4},[]")



"""Names of primitive types understand by raco.

Note that raco internally supports a limited set of types.  Different backends
can support a richer set of types, but these aren't understood by raco's
type system.  For example, raco doesn't distinguish between int and long,
so unsafe casts are not detected inside raco.
"""

# Internal and external types; these are understood by raco's type system
LONG_TYPE = "LONG_TYPE"
BOOLEAN_TYPE = "BOOLEAN_TYPE"
DOUBLE_TYPE = "DOUBLE_TYPE"
STRING_TYPE = "STRING_TYPE"
DATETIME_TYPE = "DATETIME_TYPE"

INTERNAL_TYPES = {LONG_TYPE, BOOLEAN_TYPE, DOUBLE_TYPE, STRING_TYPE, DATETIME_TYPE}  # noqa

# External only types; not understood by raco's type system
INT_TYPE = "INT_TYPE"
FLOAT_TYPE = "FLOAT_TYPE"

NUMERIC_TYPES = {LONG_TYPE, DOUBLE_TYPE}

TYPE_MAP = {k: k for k in INTERNAL_TYPES}
TYPE_MAP[INT_TYPE] = LONG_TYPE
TYPE_MAP[FLOAT_TYPE] = DOUBLE_TYPE
ALL_TYPES = TYPE_MAP.keys()


# Map from python primitive types to internal typess
python_type_map = {
    int: LONG_TYPE,
    bool: BOOLEAN_TYPE,
    float: DOUBLE_TYPE,
    str: STRING_TYPE
}

reverse_python_type_map = {v: k for k, v in python_type_map.iteritems()}


def map_type(s):
    """Convert an arbitrary type to an internal type."""
    return TYPE_MAP[s]


def parse_string(s, _type):
    """Convert from a string to an internal python representation."""
    assert _type in reverse_python_type_map
    return reverse_python_type_map[_type](s)



import abc
from raco import algebra
from raco import expression
from raco import catalog
from raco.algebra import gensym
from raco.expression import UnnamedAttributeRef
from raco.language import Language
from raco.pipelines import Pipelined

import logging
_LOG = logging.getLogger(__name__)

import re


class CodeTemplate:
    def __init__(self, s):
        self.string = s

    @classmethod
    def __format_template__(cls, s):
        """
        Format code template string
        """
        return re.sub(r'[^\S\r\n]+', ' ', s)

    def __mod__(self, other):
        return (self.__format_template__(self.string)) % other


def ct(s):
    return CodeTemplate(s)


class CBaseLanguage(Language):
    @classmethod
    def body(cls, compileResult):
        queryexec = compileResult.getExecutionCode()
        initialized = compileResult.getInitCode()
        declarations = compileResult.getDeclCode()
        resultsym = "__result__"
        return cls.base_template() % locals()

    @staticmethod
    @abc.abstractmethod
    def base_template():
        pass

    @staticmethod
    def comment(txt):
        return "// %s\n" % txt

    nextstrid = 0

    @classmethod
    def newstringident(cls):
        r = """str_%s""" % (cls.nextstrid)
        cls.nextstrid += 1
        return r

    @classmethod
    def compile_numericliteral(cls, value):
        return '%s' % (value), [], []

    @classmethod
    def negation(cls, input):
        innerexpr, decls, inits = input
        return "(!%s)" % (innerexpr,), decls, inits

    @classmethod
    def negative(cls, input):
        innerexpr, decls, inits = input
        return "(-%s)" % (innerexpr,), decls, inits

    @classmethod
    def expression_combine(cls, args, operator="&&"):
        opstr = " %s " % operator
        conjunc = opstr.join(["(%s)" % arg for arg, _, _ in args])
        decls = reduce(lambda sofar, x: sofar + x, [d for _, d, _ in args])
        inits = reduce(lambda sofar, x: sofar + x, [d for _, _, d in args])
        _LOG.debug("conjunc: %s", conjunc)
        return "( %s )" % conjunc, decls, inits

    @classmethod
    def compile_attribute(cls, expr):
        if isinstance(expr, expression.NamedAttributeRef):
            raise TypeError(
                "Error compiling attribute reference %s. \
                C compiler only support unnamed perspective. \
                Use helper function unnamed." % expr)
        if isinstance(expr, expression.UnnamedAttributeRef):
            symbol = expr.tupleref.name
            position = expr.position
            assert position >= 0
            return '%s.get(%s)' % (symbol, position), [], []


# TODO:
# The following is actually a staged materialized tuple ref.
# we should also add a staged reference tuple ref that
# just has relationsymbol and row
class StagedTupleRef:
    nextid = 0

    @classmethod
    def genname(cls):
        # use StagedTupleRef so everyone shares one mutable copy of nextid
        x = StagedTupleRef.nextid
        StagedTupleRef.nextid += 1
        return "t_%03d" % x

    def __init__(self, relsym, scheme):
        self.name = self.genname()
        self.relsym = relsym
        self.scheme = scheme
        self.__typename = None

    def getTupleTypename(self):
        if self.__typename is None:
            fields = ""
            relsym = self.relsym
            for i in range(0, len(self.scheme)):
                fieldnum = i
                fields += "_%(fieldnum)s" % locals()

            self.__typename = "MaterializedTupleRef_%(relsym)s%(fields)s" \
                              % locals()

        return self.__typename

    def generateDefinition(self):
        fielddeftemplate = """int64_t _fields[%(numfields)s];
    """
        template = """
          // can be just the necessary schema
  class %(tupletypename)s {

    public:
    %(fielddefs)s

    int64_t get(int field) const {
      return _fields[field];
    }

    void set(int field, int64_t val) {
      _fields[field] = val;
    }

    int numFields() const {
      return %(numfields)s;
    }

    %(tupletypename)s () {
      // no-op
    }

    %(tupletypename)s (std::vector<int64_t> vals) {
      for (int i=0; i<vals.size(); i++) _fields[i] = vals[i];
    }

    std::ostream& dump(std::ostream& o) const {
      o << "Materialized(";
      for (int i=0; i<numFields(); i++) {
        o << _fields[i] << ",";
      }
      o << ")";
      return o;
    }

    %(additional_code)s
  } %(after_def_code)s;
  std::ostream& operator<< (std::ostream& o, const %(tupletypename)s& t) {
    return t.dump(o);
  }

  """
        getcases = ""
        setcases = ""
        copies = ""
        numfields = len(self.scheme)
        fielddefs = fielddeftemplate % locals()

        additional_code = self.__additionalDefinitionCode__()
        after_def_code = self.__afterDefinitionCode__()

        tupletypename = self.getTupleTypename()
        relsym = self.relsym

        code = template % locals()
        return code

    def __additionalDefinitionCode__(self):
        return ""

    def __afterDefinitionCode__(self):
        return ""


def getTaggingFunc(t):
    """
    Return a visitor function that will tag
    UnnamedAttributes with the provided TupleRef
    """

    def tagAttributes(expr):
        # TODO non mutable would be nice
        if isinstance(expr, expression.UnnamedAttributeRef):
            expr.tupleref = t

        return None

    return tagAttributes


class CSelect(Pipelined, algebra.Select):
    def produce(self, state):
        self.input.produce(state)

    def consume(self, t, src, state):
        basic_select_template = """if (%(conditioncode)s) {
      %(inner_code_compiled)s
    }
    """

        condition_as_unnamed = expression.ensure_unnamed(self.condition, self)

        # tag the attributes with references
        # TODO: use an immutable approach instead
        # (ie an expression Visitor for compiling)
        [_ for _ in condition_as_unnamed.postorder(getTaggingFunc(t))]

        # compile the predicate into code
        conditioncode, cond_decls, cond_inits = \
            self.language().compile_expression(condition_as_unnamed)
        state.addInitializers(cond_inits)
        state.addDeclarations(cond_decls)

        inner_code_compiled = self.parent().consume(t, self, state)

        code = basic_select_template % locals()
        return code


class CUnionAll(Pipelined, algebra.Union):
    def produce(self, state):
        self.unifiedTupleType = self.new_tuple_ref(gensym(), self.scheme())
        state.addDeclarations([self.unifiedTupleType.generateDefinition()])

        self.right.produce(state)
        self.left.produce(state)

    def consume(self, t, src, state):
        union_template = """
        auto %(unified_tuple_name)s = \
        transpose<%(unified_tuple_typename)s>(%(src_tuple_name)s);
                        %(inner_plan_compiled)s"""

        unified_tuple_typename = self.unifiedTupleType.getTupleTypename()
        unified_tuple_name = self.unifiedTupleType.name
        src_tuple_name = t.name

        inner_plan_compiled = \
            self.parent().consume(self.unifiedTupleType, self, state)
        return union_template % locals()


class CApply(Pipelined, algebra.Apply):
    def produce(self, state):
        # declare a single new type for project
        # TODO: instead do mark used-columns?

        # always does an assignment to new tuple
        self.newtuple = self.new_tuple_ref(gensym(), self.scheme())
        state.addDeclarations([self.newtuple.generateDefinition()])

        self.input.produce(state)

    def consume(self, t, src, state):
        code = ""

        assignment_template = """
        %(dst_name)s.set(%(dst_fieldnum)s, %(src_expr_compiled)s);
        """

        dst_name = self.newtuple.name
        dst_type_name = self.newtuple.getTupleTypename()

        # declaration of tuple instance
        code += """%(dst_type_name)s %(dst_name)s;""" % locals()

        for dst_fieldnum, src_label_expr in enumerate(self.emitters):
            src_label, src_expr = src_label_expr

            # make sure to resolve attribute positions using input schema
            src_expr_unnamed = expression.ensure_unnamed(src_expr, self.input)

            # tag the attributes with references
            # TODO: use an immutable approach instead
            # (ie an expression Visitor for compiling)
            [_ for _ in src_expr_unnamed.postorder(getTaggingFunc(t))]

            src_expr_compiled, expr_decls, expr_inits = \
                self.language().compile_expression(src_expr_unnamed)
            state.addInitializers(expr_inits)
            state.addDeclarations(expr_decls)

            code += assignment_template % locals()

        innercode = self.parent().consume(self.newtuple, self, state)
        code += innercode

        return code


class CProject(Pipelined, algebra.Project):
    def produce(self, state):
        # declare a single new type for project
        # TODO: instead do mark used-columns?

        # always does an assignment to new tuple
        self.newtuple = self.new_tuple_ref(gensym(), self.scheme())
        state.addDeclarations([self.newtuple.generateDefinition()])

        self.input.produce(state)

    def consume(self, t, src, state):
        code = ""

        assignment_template = """
        %(dst_name)s.set(%(dst_fieldnum)s, %(src_name)s.get(%(src_fieldnum)s));
        """

        dst_name = self.newtuple.name
        dst_type_name = self.newtuple.getTupleTypename()
        src_name = t.name

        # declaration of tuple instance
        code += """%(dst_type_name)s %(dst_name)s;
        """ % locals()

        for dst_fieldnum, src_expr in enumerate(self.columnlist):
            if isinstance(src_expr, UnnamedAttributeRef):
                src_fieldnum = src_expr.position
            else:
                assert False, "Unsupported Project expression"
            code += assignment_template % locals()

        innercode = self.parent().consume(self.newtuple, self, state)
        code += innercode

        return code


from raco.algebra import ZeroaryOperator


class CFileScan(Pipelined, algebra.Scan):

    @abc.abstractmethod
    def __get_ascii_scan_template__(self):
        return

    @abc.abstractmethod
    def __get_binary_scan_template__(self):
        return

    def __get_relation_decl_template__(self, name):
        """
        Implement if the CFileScan implementation requires
        the relation instance to be a global declaration.
        If not then just put the local declaration within
        the *_scan_template.
        """
        return None

    def produce(self, state):

        # Common subexpression elimination
        # don't scan the same file twice
        resultsym = state.lookupExpr(self)
        _LOG.debug("lookup %s(h=%s) => %s", self, self.__hash__(), resultsym)
        if not resultsym:
            # TODO for now this will break
            # whatever relies on self.bound like reusescans
            # Scan is the only place where a relation is declared
            resultsym = gensym()

            name = str(self.relation_key).split(':')[2]
            fscode = self.__compileme__(resultsym, name)
            state.saveExpr(self, resultsym)

            stagedTuple = self.new_tuple_ref(resultsym, self.scheme())
            state.saveTupleDef(resultsym, stagedTuple)

            tuple_type_def = stagedTuple.generateDefinition()
            tuple_type = stagedTuple.getTupleTypename()
            state.addDeclarations([tuple_type_def])

            rel_decl_template = self.__get_relation_decl_template__(name)
            if rel_decl_template:
                state.addDeclarations([rel_decl_template % locals()])

            # now that we have the type, format this in;
            state.setPipelineProperty('type', 'scan')
            state.setPipelineProperty('source', self.__class__)
            state.addPipeline(fscode % {"result_type": tuple_type})

        # no return value used because parent is a new pipeline
        self.parent().consume(resultsym, self, state)

    def consume(self, t, src, state):
        assert False, "as a source, no need for consume"

    def __compileme__(self, resultsym, name):
        # TODO use the identifiers (don't split str and extract)
        # name = self.relation_key

        _LOG.debug('compiling file scan for relation_key %s'
                   % self.relation_key)

        # tup = (resultsym, self.originalterm.originalorder, self.originalterm)
        # self.trace("// Original query position of %s: term %s (%s)" % tup)

        if isinstance(self.relation_key, catalog.ASCIIFile):
            code = self.__get_ascii_scan_template__() % locals()
        else:
            code = self.__get_binary_scan_template__() % locals()
        return code

    def __str__(self):
        return "%s(%s)" % (self.opname(), self.relation_key)

    def __eq__(self, other):
        """
        For what we are using FileScan for, the only use
        of __eq__ is in hashtable lookups for CSE optimization.
        We omit self.schema because the relation_key determines
        the level of equality needed.

        This could break other things, so better may be to
        make a normalized copy of an expression. This could
        include simplification but in the case of Scans make
        the scheme more generic.

        @see MemoryScan.__eq__
        """
        return ZeroaryOperator.__eq__(self, other) and \
            self.relation_key == other.relation_key


# Rules
from raco import rules


class BreakHashJoinConjunction(rules.Rule):
    """A rewrite rule for turning HashJoin(a=c and b=d)
    into select(b=d)[HashJoin(a=c)]"""

    def __init__(self, select_clazz, join_clazz):
        self.select_clazz = select_clazz
        self.join_clazz = join_clazz

    def fire(self, expr):
        if isinstance(expr, self.join_clazz) \
                and isinstance(expr.condition.left, expression.EQ) \
                and isinstance(expr.condition.right, expression.EQ):
            return self.select_clazz(expr.condition.right,
                                     self.join_clazz(expr.condition.left,
                                                     expr.left,
                                                     expr.right))

        return expr

    def __str__(self):
        return "%s(a=c and b=d) => %s(b=d)[%s(a=c)]" \
               % (self.join_clazz.__name__,
                  self.select_clazz.__name__,
                  self.join_clazz.__name__)


clang_push_select = [
    rules.SplitSelects(),
    rules.PushSelects(),
    # We don't want to merge selects because it doesn't really
    # help and it (maybe) creates HashJoin(conjunction)
    # rules.MergeSelects()
]

EMIT_CONSOLE = 'console'
EMIT_FILE = 'file'


class BaseCStore(Pipelined, algebra.Store):
    def __init__(self, emit_print, relation_key, plan):
        super(BaseCStore, self).__init__(relation_key, plan)
        self.emit_print = emit_print

    def produce(self, state):
        self.input.produce(state)

    def consume(self, t, src, state):
        code = ""
        resdecl = "std::vector<%s> result;\n" % (t.getTupleTypename())
        state.addDeclarations([resdecl])
        code += "result.push_back(%s);\n" % (t.name)

        if self.emit_print == EMIT_CONSOLE:
            code += self.language().log_unquoted("%s" % t.name, 2)
        elif self.emit_print == EMIT_FILE:
            code += self.__file_code__(t, state)

        return code

    @abc.abstractmethod
    def __file_code__(self, t, state):
        pass

    def __repr__(self):
        return "{op}({ep!r}, {rk!r}, {pl!r})".format(op=self.opname(),
                                                     ep=self.emit_print,
                                                     rk=self.relation_key,
                                                     pl=self.input)


class StoreToBaseCStore(rules.Rule):
    """A rule to store tuples into emit_print"""
    def __init__(self, emit_print, subclass):
        self.emit_print = emit_print
        assert issubclass(subclass, BaseCStore), \
            "%s is not a subclass of %s" % (subclass, BaseCStore)
        self.subclass = subclass

    def fire(self, expr):
        if isinstance(expr, algebra.Store):
            return self.subclass(self.emit_print,
                                 expr.relation_key,
                                 expr.input)
        return expr

    def __str__(self):
        return "Store => %s" % self.subclass.__name__



"""Various tests of type safety."""
import unittest

from raco.fakedb import FakeDatabase
from raco.scheme import Scheme
from raco.myrial.myrial_test import MyrialTestCase
from raco.expression import TypeSafetyViolation
from collections import Counter
from raco import types


class TypeTests(MyrialTestCase):
    schema = Scheme(
        [("clong", types.LONG_TYPE),
         ("cint", types.INT_TYPE),
         ("cstring", types.STRING_TYPE),
         ("cfloat", types.DOUBLE_TYPE),
         ("cbool", types.BOOLEAN_TYPE),
         ("cdate", types.DATETIME_TYPE)])

    def setUp(self):
        super(TypeTests, self).setUp()
        self.db.ingest("public:adhoc:mytable", Counter(), TypeTests.schema)

    def test_noop(self):
        query = """
        X = SCAN(public:adhoc:mytable);
        STORE(X, OUTPUT);
        """

        self.check_scheme(query, TypeTests.schema)

    def test_invalid_eq1(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS X EMIT clong=cstring];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_eq2(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS X EMIT cfloat=cdate];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_ge(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS X EMIT cfloat>=cstring];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_lt(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS X EMIT cfloat<cdate];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_and(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS X EMIT cfloat AND cdate];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_or(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS X EMIT cfloat OR cdate];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_not(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS X EMIT not cdate];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_plus(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS X EMIT cdate + cint];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_times(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS X EMIT cdate * cstring];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_divide(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS X EMIT cdate / clong];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_divide(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS X EMIT cfloat / cfloat AS y];
        STORE(X, OUTPUT);
        """
        schema = Scheme([('y', types.DOUBLE_TYPE)])
        self.check_scheme(query, schema)

    def test_idivide(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS X EMIT cfloat // cint AS y];
        STORE(X, OUTPUT);
        """
        schema = Scheme([('y', types.LONG_TYPE)])
        self.check_scheme(query, schema)

    def test_neg(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS X EMIT -cstring];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_case1(self):
        query = """
        t = SCAN(public:adhoc:mytable);
        rich = [FROM t EMIT
                CASE WHEN clong <= 5000 THEN "poor"
                     WHEN clong <= 25000 THEN 5.5
                     ELSE "rich" END];
        STORE(rich, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_case2(self):
        query = """
        t = SCAN(public:adhoc:mytable);
        rich = [FROM t EMIT
                CASE WHEN clong <= 5000 THEN "poor"
                     WHEN clong <= 25000 THEN "middle class"
                     ELSE 1922 END];
        STORE(rich, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_cos(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS Y
             WHERE cos(cbool) > 1.0 EMIT *];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_floor(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS Y
             WHERE floor(cbool) > 1.0 EMIT *];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_ceil(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS Y
             WHERE ceil(clong) == "Hello world" EMIT *];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_tan(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS Y
             WHERE tan(cstring) > 1.0 EMIT *];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_pow(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS X EMIT POW(cfloat, cstring)];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_substr(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS X EMIT SUBSTR(0, 3, cfloat)];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_len(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS X EMIT LEN(cfloat)];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_greater(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS X EMIT GREATER(cbool, 3)];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_lesser(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS X EMIT LESSER(3.0, cdate)];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_sum_compare(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS X EMIT SUM(clong) > "foo"];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_max_compare(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS X EMIT MAX(cstring) == 7];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_avg(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS X EMIT AVG(cbool)];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)

    def test_invalid_stdev(self):
        query = """
        X = [FROM SCAN(public:adhoc:mytable) AS X EMIT STDEV(cdate)];
        STORE(X, OUTPUT);
        """
        with self.assertRaises(TypeSafetyViolation):
            self.check_scheme(query, None)



from raco.datalog.grammar import parse
from raco.compile import optimize

import logging
LOG = logging.getLogger(__name__)


class RACompiler(object):
    """Thin wrapper interface for lower level functions parse, optimize,
    compile"""

    def fromDatalog(self, program):
        """Parse datalog and convert to RA"""
        self.physicalplan = None
        self.source = program
        self.parsed = parse(program)
        LOG.debug("parser output: %s", self.parsed)
        self.logicalplan = self.parsed.toRA()

    def optimize(self, target, **kwargs):
        """Convert logical plan to physical plan"""
        self.physicalplan = optimize(self.logicalplan, target, **kwargs)


import raco.rules as rules
from raco.language import Algebra


class OptLogicalAlgebra(Algebra):
    @staticmethod
    def opt_rules(**kwargs):
        return [rules.RemoveTrivialSequences(),
                rules.SimpleGroupBy(),
                rules.SplitSelects(),
                rules.PushSelects(),
                rules.MergeSelects(),
                rules.ProjectToDistinctColumnSelect(),
                rules.JoinToProjectingJoin(),
                rules.PushApply(),
                rules.RemoveUnusedColumns(),
                rules.PushApply(),
                rules.RemoveUnusedColumns(),
                rules.PushApply()]


# TODO: To be refactored into shared memory lang,
# where you plugin in the sequential shared memory language specific codegen

from raco import algebra
from raco import expression
from raco.language import clangcommon, Algebra
from raco import rules
from raco.pipelines import Pipelined
from raco.language.clangcommon import StagedTupleRef, ct, CBaseLanguage

from raco.algebra import gensym

import logging

_LOG = logging.getLogger(__name__)

import itertools
import os.path


template_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             "c_templates")


def readtemplate(fname):
    return file(os.path.join(template_path, fname)).read()


template_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             "c_templates")

base_template = readtemplate("base_query.template")
twopass_select_template = readtemplate("precount_select.template")
hashjoin_template = readtemplate("hashjoin.template")
filteringhashjoin_template = ""
filtering_nestedloop_join_chain_template = ""
# =readtemplate("filtering_nestedloop_join_chain.template")
ascii_scan_template = readtemplate("ascii_scan.template")
binary_scan_template = readtemplate("binary_scan.template")


class CStagedTupleRef(StagedTupleRef):
    def __additionalDefinitionCode__(self):
        constructor_template = """
    public:
    %(tupletypename)s (relationInfo * rel, int row) {
      %(copies)s
    }
    """

        copytemplate = """_fields[%(fieldnum)s] = \
        rel->relation[row*rel->fields + %(fieldnum)s];
    """

        copies = ""
        # TODO: actually list the trimmed schema offsets
        for i in range(0, len(self.scheme)):
            fieldnum = i
            copies += copytemplate % locals()

        tupletypename = self.getTupleTypename()
        return constructor_template % locals()


class CC(CBaseLanguage):
    @staticmethod
    def base_template():
        return base_template

    @staticmethod
    def pipeline_wrap(ident, code, attrs):

        # timing code
        if True:
            inner_code = code
            timing_template = ct("""auto start_%(ident)s = walltime();
            %(inner_code)s
            auto end_%(ident)s = walltime();
            auto runtime_%(ident)s = end_%(ident)s - start_%(ident)s;
            std::cout << "pipeline %(ident)s: " << runtime_%(ident)s \
                        << " s" \
                        << std::endl;
            std::cout << "timestamp %(ident)s start " \
                        << std::setprecision(15) \
                        << start_%(ident)s << std::endl;
            std::cout << "timestamp %(ident)s end " \
                        << std::setprecision(15) \
                        << end_%(ident)s << std::endl;
            """)
            code = timing_template % locals()

        return code

    @staticmethod
    def group_wrap(ident, grpcode, attrs):
        pipeline_template = ct("""
        auto start_%(ident)s = walltime();
        %(grpcode)s
        auto end_%(ident)s = walltime();
        auto runtime_%(ident)s = end_%(ident)s - start_%(ident)s;
        std::cout << "pipeline group %(ident)s: " \
                    << runtime_%(ident)s \
                    << " s" << std::endl;
        """)

        code = pipeline_template % locals()
        return code

    @staticmethod
    def log(txt):
        return """std::cout << "%s" << std::endl;
        """ % txt

    @staticmethod
    def log_unquoted(code, level=0):
        return """std::cout << %s << std::endl;
      """ % code

    @staticmethod
    def log_file(code, level=0):
        return """logfile << "%s" << "\\n";\n """ % code

    @staticmethod
    def log_file_unquoted(code, level=0):
        return """logfile << %s << " ";\n """ % code

    @classmethod
    def compile_stringliteral(cls, s):
        sid = cls.newstringident()
        lookup_init = """auto %s = string_index.string_lookup(%s);""" \
                      % (sid, s)
        build_init = """
        string_index = build_string_index("sp2bench_1m.index");
        """
        return """(%s)""" % sid, [], [build_init, lookup_init]
        # raise ValueError("String Literals not supported\
        # in C language: %s" % s)


class CCOperator(Pipelined, algebra.Operator):
    _language = CC

    @classmethod
    def new_tuple_ref(cls, sym, scheme):
        return CStagedTupleRef(sym, scheme)

    @classmethod
    def language(cls):
        return cls._language

    def postorder_traversal(self, func):
        return self.postorder(func)


from raco.algebra import UnaryOperator


class CMemoryScan(algebra.UnaryOperator, CCOperator):
    def produce(self, state):
        self.input.produce(state)

    # TODO: when have pipeline tree representation,
    # TODO: will have a consumeMaterialized() method instead;
    # TODO: for now we reuse the tuple-based consume
    def consume(self, inputsym, src, state):
        # now generate the scan from memory

        # TODO: generate row variable to avoid naming conflict for nested scans
        memory_scan_template = """for (uint64_t i : %(inputsym)s->range()) {
          %(tuple_type)s %(tuple_name)s(%(inputsym)s, i);

          %(inner_plan_compiled)s
       } // end scan over %(inputsym)s
       """

        stagedTuple = state.lookupTupleDef(inputsym)
        tuple_type = stagedTuple.getTupleTypename()
        tuple_name = stagedTuple.name

        inner_plan_compiled = self.parent().consume(stagedTuple, self, state)

        code = memory_scan_template % locals()
        state.setPipelineProperty("type", "in_memory")
        state.addPipeline(code)
        return None

    def num_tuples(self):
        raise NotImplementedError("{}.num_tuples()".format(op=self.opname()))

    def shortStr(self):
        return "%s" % (self.opname())

    def __eq__(self, other):
        """
    For what we are using MemoryScan for, the only use
    of __eq__ is in hashtable lookups for CSE optimization.
    We omit self.schema because the relation_key determines
    the level of equality needed.

    @see FileScan.__eq__
    """
        return UnaryOperator.__eq__(self, other)


class CGroupBy(algebra.GroupBy, CCOperator):
    _i = 0

    @classmethod
    def __genHashName__(cls):
        name = "group_hash_%03d" % cls._i
        cls._i += 1
        return name

    def produce(self, state):
        assert len(self.grouping_list) <= 2, \
            "%s does not currently support groupings of \
            more than 2 attributes" % self.__class__.__name__
        assert len(self.aggregate_list) == 1, \
            """%s currently only supports aggregates of 1 attribute
            (aggregate_list=%s)""" \
            % (self.__class__.__name__, self.aggregate_list)
        for agg_term in self.aggregate_list:
            assert isinstance(agg_term,
                              expression.BuiltinAggregateExpression), \
                """%s only supports simple aggregate expressions.
                A rule should create Apply[GroupBy]""" \
                % self.__class__.__name__

        self.useMap = len(self.grouping_list) > 0

        if self.useMap:
            if len(self.grouping_list) == 1:
                declr_template = """std::unordered_map<int64_t, int64_t> \
                %(hashname)s;
          """
            elif len(self.grouping_list) == 2:
                declr_template = """std::unordered_map<\
                std::pair<int64_t, int64_t>, int64_t, pairhash> \
                %(hashname)s;
                """
        else:
            declr_template = """int64_t %(hashname)s;
            """

        self.hashname = self.__genHashName__()
        hashname = self.hashname

        hash_declr = declr_template % locals()
        state.addDeclarations([hash_declr])

        my_sch = self.scheme()

        _LOG.debug("aggregates: %s", self.aggregate_list)
        _LOG.debug("columns: %s", self.column_list())
        _LOG.debug("groupings: %s", self.grouping_list)
        _LOG.debug("groupby scheme: %s", my_sch)
        _LOG.debug("groupby scheme[0] type: %s", type(my_sch[0]))

        self.input.produce(state)

        # now that everything is aggregated, produce the tuples
        assert (not self.useMap) \
            or isinstance(self.column_list()[0],
                          expression.AttributeRef), \
            "assumes first column is the key and " \
            "second is aggregate result: %s" % (self.column_list()[0])

        if self.useMap:
            if len(self.grouping_list) == 1:
                produce_template = """for (auto it=%(hashname)s.begin(); \
                it!=%(hashname)s.end(); it++) {
                %(output_tuple_type)s %(output_tuple_name)s(\
                {it->first, it->second});
                %(inner_code)s
                }
                """
            elif len(self.grouping_list) == 2:
                produce_template = """for (auto it=%(hashname)s.begin(); \
                it!=%(hashname)s.end(); it++) {
                %(output_tuple_type)s %(output_tuple_name)s(\
                {it->first.first, it->first.second, it->second});
                %(inner_code)s
                }
                """
        else:
            produce_template = """{
            %(output_tuple_type)s %(output_tuple_name)s({ %(hashname)s });
            %(inner_code)s
            }
            """

        output_tuple = CStagedTupleRef(gensym(), my_sch)
        output_tuple_name = output_tuple.name
        output_tuple_type = output_tuple.getTupleTypename()
        state.addDeclarations([output_tuple.generateDefinition()])

        inner_code = self.parent().consume(output_tuple, self, state)
        code = produce_template % locals()
        state.setPipelineProperty("type", "in_memory")
        state.addPipeline(code)

    def consume(self, inputTuple, fromOp, state):
        if self.useMap:
            if len(self.grouping_list) == 1:
                materialize_template = """%(op)s_insert(%(hashname)s, \
                %(tuple_name)s, %(key1pos)s, %(valpos)s);
                """
            elif len(self.grouping_list) == 2:
                materialize_template = """%(op)s_insert(%(hashname)s, \
                %(tuple_name)s, %(key1pos)s, %(key2pos)s, %(valpos)s);
                """
        else:
            materialize_template = """%(op)s_insert(%(hashname)s, \
            %(tuple_name)s, %(valpos)s);
            """

        hashname = self.hashname
        tuple_name = inputTuple.name

        # make key from grouped attributes
        if self.useMap:
            inp_sch = self.input.scheme()

            key1pos = self.grouping_list[0].get_position(inp_sch)

            if len(self.grouping_list) == 2:
                key2pos = self.grouping_list[1].get_position(
                    inp_sch)

        if isinstance(self.aggregate_list[0], expression.ZeroaryOperator):
            # no value needed for Zero-input aggregate,
            # but just provide the first column
            valpos = 0
        elif isinstance(self.aggregate_list[0], expression.UnaryOperator):
            # get value positions from aggregated attributes
            valpos = self.aggregate_list[0].input.get_position(self.scheme())
        else:
            assert False, "only support Unary or Zeroary aggregates"

        op = self.aggregate_list[0].__class__.__name__

        code = materialize_template % locals()
        return code


class CHashJoin(algebra.Join, CCOperator):
    _i = 0

    @classmethod
    def __genHashName__(cls):
        name = "hash_%03d" % cls._i
        cls._i += 1
        return name

    def produce(self, state):
        if not isinstance(self.condition, expression.EQ):
            msg = "The C compiler can only handle equi-join conditions of \
            a single attribute: %s" % self.condition
            raise ValueError(msg)

        left_sch = self.left.scheme()

        # find the attribute that corresponds to the right child
        self.rightCondIsRightAttr = \
            self.condition.right.position >= len(left_sch)
        self.leftCondIsRightAttr = \
            self.condition.left.position >= len(left_sch)
        assert self.rightCondIsRightAttr ^ self.leftCondIsRightAttr

        # find the attribute that corresponds to the right child
        if self.rightCondIsRightAttr:
            self.right_keypos = \
                self.condition.right.position - len(left_sch)
        else:
            self.right_keypos = \
                self.condition.left.position - len(left_sch)

        # find the attribute that corresponds to the left child
        if self.rightCondIsRightAttr:
            self.left_keypos = self.condition.left.position
        else:
            self.left_keypos = self.condition.right.position

        self.right.childtag = "right"
        # common index is defined by same right side and same key
        hashsym = state.lookupExpr((self.right, self.right_keypos))

        if not hashsym:
            # if right child never bound then store hashtable symbol and
            # call right child produce
            self._hashname = self.__genHashName__()
            _LOG.debug("generate hashname %s for %s", self._hashname, self)
            state.saveExpr((self.right, self.right_keypos), self._hashname)
            self.right.produce(state)
        else:
            # if found a common subexpression on right child then
            # use the same hashtable
            self._hashname = hashsym
            _LOG.debug("reuse hash %s for %s", self._hashname, self)

        self.left.childtag = "left"
        self.left.produce(state)

    def consume(self, t, src, state):
        if src.childtag == "right":
            declr_template = """std::unordered_map\
            <int64_t, std::vector<%(in_tuple_type)s>* > %(hashname)s;
            """

            right_template = """insert(%(hashname)s, %(keyname)s, %(keypos)s);
            """

            hashname = self._hashname
            keyname = t.name

            keypos = self.right_keypos

            in_tuple_type = t.getTupleTypename()

            # declaration of hash map
            hashdeclr = declr_template % locals()
            state.addDeclarations([hashdeclr])

            # materialization point
            code = right_template % locals()

            return code

        if src.childtag == "left":
            left_template = """
          for (auto %(right_tuple_name)s : \
          lookup(%(hashname)s, %(keyname)s.get(%(keypos)s))) {
            auto %(out_tuple_name)s = \
            combine<%(out_tuple_type)s> (%(keyname)s, %(right_tuple_name)s);
         %(inner_plan_compiled)s
      }
      """
            hashname = self._hashname
            keyname = t.name
            keytype = t.getTupleTypename()

            keypos = self.left_keypos

            right_tuple_name = gensym()

            outTuple = CStagedTupleRef(gensym(), self.scheme())
            out_tuple_type_def = outTuple.generateDefinition()
            out_tuple_type = outTuple.getTupleTypename()
            out_tuple_name = outTuple.name

            state.addDeclarations([out_tuple_type_def])

            inner_plan_compiled = self.parent().consume(outTuple, self, state)

            code = left_template % locals()
            return code

        assert False, "src not equal to left or right"


def indentby(code, level):
    indent = " " * ((level + 1) * 6)
    return "\n".join([indent + line for line in code.split("\n")])


# iteration  over table + insertion into hash table with filter

class CUnionAll(clangcommon.CUnionAll, CCOperator):
    pass


class CApply(clangcommon.CApply, CCOperator):
    pass


class CProject(clangcommon.CProject, CCOperator):
    pass


class CSelect(clangcommon.CSelect, CCOperator):
    pass


class CFileScan(clangcommon.CFileScan, CCOperator):
    def __get_ascii_scan_template__(self):
        return ascii_scan_template

    def __get_binary_scan_template__(self):
        return binary_scan_template


class CStore(clangcommon.BaseCStore, CCOperator):

    def __file_code__(self, t, state):
        code = ""
        state.addPreCode('std::ofstream logfile;\n')
        resultfile = str(self.relation_key).split(":")[2]
        opentuple = 'logfile.open("%s");\n' % resultfile
        schemafile = self.write_schema(self.scheme())
        state.addPreCode(schemafile)
        state.addPreCode(opentuple)
        code += "int logi = 0;\n"
        code += "for (logi = 0; logi < %s.numFields() - 1; logi++) {\n" \
                % (t.name)
        code += self.language().log_file_unquoted("%s.get(logi)" % t.name)
        code += "}\n "
        code += "logfile << %s.get(logi);\n" % (t.name)
        code += "logfile << '\\n';"
        state.addPostCode('logfile.close();')

        return code

    def write_schema(self, scheme):
        schemafile = 'schema/' + str(self.relation_key).split(":")[2]
        code = 'logfile.open("%s");\n' % schemafile
        names = [x.encode('UTF8') for x in scheme.get_names()]
        code += self.language().log_file("%s" % names)
        code += self.language().log_file("%s" % scheme.get_types())
        code += 'logfile.close();'
        return code


class MemoryScanOfFileScan(rules.Rule):
    """A rewrite rule for making a scan into
    materialization in memory then memory scan"""

    def fire(self, expr):
        if isinstance(expr, algebra.Scan) and not isinstance(expr, CFileScan):
            return CMemoryScan(CFileScan(expr.relation_key, expr.scheme()))
        return expr

    def __str__(self):
        return "Scan => MemoryScan[FileScan]"


def clangify(emit_print):
    return [
        rules.ProjectingJoinToProjectOfJoin(),

        rules.OneToOne(algebra.Select, CSelect),
        MemoryScanOfFileScan(),
        rules.OneToOne(algebra.Apply, CApply),
        rules.OneToOne(algebra.Join, CHashJoin),
        rules.OneToOne(algebra.GroupBy, CGroupBy),
        rules.OneToOne(algebra.Project, CProject),
        rules.OneToOne(algebra.UnionAll, CUnionAll),
        # TODO: obviously breaks semantics
        rules.OneToOne(algebra.Union, CUnionAll),
        clangcommon.StoreToBaseCStore(emit_print, CStore),

        clangcommon.BreakHashJoinConjunction(CSelect, CHashJoin)
    ]


class CCAlgebra(Algebra):
    def __init__(self, emit_print=clangcommon.EMIT_CONSOLE):
        """ To store results into a file or onto console """
        self.emit_print = emit_print

    def opt_rules(self, **kwargs):
        # Sequence that works for datalog
        # TODO: replace with below
        # datalog_rules = [
        # rules.CrossProduct2Join(),
        # rules.SimpleGroupBy(),
        # rules.OneToOne(algebra.Select, CSelect),
        # MemoryScanOfFileScan(),
        # rules.OneToOne(algebra.Apply, CApply),
        # rules.OneToOne(algebra.Join, CHashJoin),
        # rules.OneToOne(algebra.GroupBy, CGroupBy),
        # rules.OneToOne(algebra.Project, CProject),
        # TODO: obviously breaks semantics
        # rules.OneToOne(algebra.Union, CUnionAll),
        # rules.FreeMemory()
        # ]

        # sequence that works for myrial
        rule_grps_sequence = [
            rules.remove_trivial_sequences,
            rules.simple_group_by,
            clangcommon.clang_push_select,
            [rules.ProjectToDistinctColumnSelect(),
             rules.JoinToProjectingJoin()],
            rules.push_apply,
            clangify(self.emit_print)
        ]

        if kwargs.get('SwapJoinSides'):
            rule_grps_sequence.insert(0, [rules.SwapJoinSides()])

        return list(itertools.chain(*rule_grps_sequence))



"""Test of Myrial's control flow graph generation."""

import collections

import raco.myrial.myrial_test as myrial_test
import raco.scheme as scheme
from raco import types
import networkx as nx


class CFGTest(myrial_test.MyrialTestCase):
    points_table = collections.Counter()
    points_schema = scheme.Scheme([('id', types.LONG_TYPE),
                                   ('x', types.DOUBLE_TYPE),
                                   ('y', types.DOUBLE_TYPE)])
    points_key = "public:adhoc:points"

    def setUp(self):
        super(CFGTest, self).setUp()

        self.db.ingest(CFGTest.points_key,
                       CFGTest.points_table,
                       CFGTest.points_schema)

    def test_cfg(self):
        query = """
        Point = SCAN(public:adhoc:points);

        DO
          Big = [FROM Point WHERE x * y > 100 EMIT *];
          Continue = [FROM Big, Point EMIT COUNT(*) > 0 AS cnt];
        WHILE Continue;

        STORE(Big, OUTPUT);
        """

        statements = self.parser.parse(query)
        self.processor.evaluate(statements)

        expected = nx.DiGraph()
        expected.add_node(0, def_var="Point", uses=set())
        expected.add_node(1, def_var="Big", uses={"Point"})
        expected.add_node(2, def_var="Continue", uses={"Big", "Point"})
        expected.add_node(3, def_var=None, uses={"Continue"})
        expected.add_node(4, def_var=None, uses={"Big"})

        for i in range(4):
            expected.add_edge(i, i + 1)
        expected.add_edge(3, 1)

        actual = self.processor.cfg.graph

        self.assertEquals(actual.adj, expected.adj)
        self.assertEquals(len(actual), len(expected))

        for n in expected:
            self.assertIn(n, actual)
            self.assertEquals(actual.node[n]['uses'], expected.node[n]['uses'])
            self.assertEquals(actual.node[n]['def_var'],
                              expected.node[n]['def_var'])

        live_in, live_out = self.processor.cfg.compute_liveness()

        self.assertEquals(live_out, {0: {'Point'}, 1: {'Point', 'Big'},
                                     2: {'Continue', 'Big', 'Point'},
                                     3: {'Big', 'Point'}, 4: set()})

        self.assertEquals(live_in, {0: set(), 1: {'Point'},
                                    2: {'Big', 'Point'},
                                    3: {'Big', 'Point', 'Continue'},
                                    4: {'Big'}})

    def test_dead_code_elim(self):
        with open('examples/deadcode.myl') as fh:
            query = fh.read()

        statements = self.parser.parse(query)
        self.processor.evaluate(statements)
        self.assertEquals(set(self.processor.cfg.graph.nodes()), set(range(9)))

        self.processor.cfg.dead_code_elimination()
        self.assertEquals(set(self.processor.cfg.graph.nodes()), {2, 6, 7, 8})

    def test_bug_245_dead_loop_elim_do_while(self):
        with open('examples/deadcode2.myl') as fh:
            query = fh.read()

        statements = self.parser.parse(query)
        self.processor.evaluate(statements)
        self.assertEquals(set(self.processor.cfg.graph.nodes()), set(range(3)))

        self.processor.cfg.dead_loop_elimination()
        self.processor.cfg.dead_code_elimination()
        self.assertEquals(set(self.processor.cfg.graph.nodes()), set())

    def test_dead_loop_interior(self):
        """Test of a dead loop before the end of the program."""
        query = """
        x = [0 as val, 1 as exp];
        y = x;

        do
            x = [from x emit val+1 as val, 2*exp as exp];
        while [from x emit val < 5];
        store(y, OUTPUT);
        """

        statements = self.parser.parse(query)
        self.processor.evaluate(statements)
        self.assertEquals(set(self.processor.cfg.graph.nodes()), set(range(5)))

        self.processor.cfg.dead_loop_elimination()
        self.processor.cfg.dead_code_elimination()

        self.assertEquals(set(self.processor.cfg.graph.nodes()), {0, 1, 4})

    def test_two_dead_loops(self):
        """Test of two unrelated dead loops."""
        query = """
        x = [0 as val, 1 as exp];
        y = x;
        z = y;

        do
            x = [from x emit val+1 as val, 2*exp as exp];
        while [from x emit val < 5];
        do
            z = [from z emit val+1 as val, 2*exp as exp];
        while [from z emit val < 5];
        store(y, OUTPUT);
        """

        statements = self.parser.parse(query)
        self.processor.evaluate(statements)
        self.assertEquals(set(self.processor.cfg.graph.nodes()), set(range(8)))

        self.processor.cfg.dead_loop_elimination()
        self.processor.cfg.dead_code_elimination()
        self.assertEquals(set(self.processor.cfg.graph.nodes()), {0, 1, 7})

    def test_two_dead_loops_samevar(self):
        """Test that recursive calls to dead_loop_elimination remove
        repeated dead loops reading/writing the same variable."""
        query = """
        x = [0 as val, 1 as exp];
        y = x;

        do
            x = [from x emit val+1 as val, 2*exp as exp];
        while [from x emit val < 5];

        do
            x = [from x emit val+1 as val, 2*exp as exp];
        while [from x emit val < 5];

        store(y, OUTPUT);
        """

        statements = self.parser.parse(query)
        self.processor.evaluate(statements)
        self.assertEquals(set(self.processor.cfg.graph.nodes()), set(range(7)))

        self.processor.cfg.dead_loop_elimination()
        self.processor.cfg.dead_code_elimination()
        self.assertEquals(set(self.processor.cfg.graph.nodes()), {0, 1, 6})

    def test_chaining(self):
        query = """
        A = SCAN(public:adhoc:points);
        B = SCAN(public:adhoc:points);
        C = UNIONALL(A, B);
        D = DISTINCT(C);
        E = SCAN(public:adhoc:points);
        F = DIFF(E, D);
        G = DISTINCT(F);
        STORE(G, OUTPUT);
        """

        statements = self.parser.parse(query)
        self.processor.evaluate(statements)
        self.assertEquals(len(self.processor.cfg.graph), 8)

        self.processor.cfg.apply_chaining()
        self.assertEquals(len(self.processor.cfg.graph), 1)

    def test_chaining_variable_reuse(self):
        """Test of chaining with re-used variable names."""
        query = """
        X = SCAN(public:adhoc:points);
        Y = SCAN(public:adhoc:points);
        X = [FROM X, Y WHERE X.x == Y.y EMIT Y.*];
        X = DISTINCT(X);
        STORE(X, OUTPUT);
        """
        statements = self.parser.parse(query)
        self.processor.evaluate(statements)
        self.assertEquals(len(self.processor.cfg.graph), 5)

        self.processor.cfg.apply_chaining()
        self.assertEquals(self.processor.cfg.graph.nodes(), [4])
        self.assertEquals(len(self.processor.cfg.graph.node[4]['uses']), 0)

    def test_chaining_dead_code_elim(self):
        query = """
        Q = DISTINCT(SCAN(public:adhoc:points));
        DO
            A = SCAN(public:adhoc:points);
            B = SCAN(public:adhoc:points);
            P = DISTINCT(A);
            C = DIFF(A, B);
            Continue = COUNTALL(C);
        WHILE Continue;
        STORE(C, OUTPUT);
        X = SCAN(public:adhoc:points);
        """

        statements = self.parser.parse(query)
        self.processor.evaluate(statements)
        self.assertEquals(len(self.processor.cfg.graph), 9)

        self.processor.cfg.dead_code_elimination()
        self.assertEquals(set(self.processor.cfg.graph.nodes()),
                          {1, 2, 4, 5, 6, 7})

        self.processor.cfg.apply_chaining()
        self.assertEquals(set(self.processor.cfg.graph.nodes()), {4, 6, 7})



# These imports are required here -- for eval inside replace_with_repr
from raco.expression import *
from raco.algebra import *
from raco.relation_key import *
from raco.scheme import *
from raco.language.myrialang import *
from raco.language.clang import *
from raco.language.grappalang import *

# NOTES: relying on import * for eval is error prone due
#        to namespace collisions
# NOTES: what to do if a operator has two constructors?


def replace_with_repr(plan):
    r = repr(plan)
    try:
        return eval(r)
    except (TypeError, AttributeError, SyntaxError):
        print 'Error with repr {r} of plan {p}'.format(r=r, p=plan)
        raise


from raco import expression
from raco import scheme
from raco.utility import Printable, real_str

from abc import ABCMeta, abstractmethod
import copy
import operator
import math
from raco.expression import StateVar


# BEGIN Code to generate variables names
var_id = 0


def reset():
    global var_id
    var_id = 0


def gensym():
    global var_id
    var_id += 1
    return "V%s" % var_id
# END Code to generate variables names


# Global constants
DEFAULT_CARDINALITY = 10000


class RecursionError(ValueError):
    pass


class SchemaError(Exception):
    pass


class Operator(Printable):
    """Operator base class"""
    __metaclass__ = ABCMeta

    def __init__(self):
        self.bound = None
        # Extra code to emit to cleanup
        self.cleanup = ""
        self.alias = self
        self._trace = []

    @abstractmethod
    def apply(self, f):
        """ apply function f to its children. """

    @abstractmethod
    def children(self):
        """Return all the children of this operator."""

    @abstractmethod
    def scheme(self):
        """Return the scheme of the tuples output by this operator."""

    def walk(self):
        """Return an iterator over the tree of operators."""
        yield self
        for c in self.children():
            for x in c.walk():
                yield x

    @abstractmethod
    def num_tuples(self):
        """Return the expected number of tuples output by this operator."""

    def postorder(self, f):
        """Postorder traversal, applying a function to each operator.  The
        function returns an iterator"""
        for c in self.children():
            for x in c.postorder(f):
                yield x
        for x in f(self):
            yield x

    def preorder(self, f):
        """Preorder traversal, applying a function to each operator.  The
        function returns an iterator"""
        for x in f(self):
            yield x
        for c in self.children():
            for x in c.postorder(f):
                yield x

    def collectParents(self, parent_map=None):
        """Construct a dict mapping children to parents. Used in
        optimization"""
        if parent_map is None:
            parent_map = {}
        for c in self.children():
            parent_map.setdefault(c, []).append(self)
            c.collectParents(parent_map)

    def __copy__(self):
        raise RuntimeError("Shallow copy not supported for operators")

    def __eq__(self, other):
        return self.__class__ == other.__class__

    def __str__(self):
        if len(self.children()) > 0:
            return "%s%s" % (self.shortStr(), real_str(self.children()))
        return self.shortStr()

    def __hash__(self):
        h = str(self.__class__).__hash__()
        for i, c in enumerate(self.children()):
            h ^= (c.__hash__() << i)

        return h

    def copy(self, other):
        self._trace = [pair for pair in other.gettrace()]
        self.bound = None

    def trace(self, key, val):
        self._trace.append((key, val))

    def gettrace(self):
        """Return a list of trace messages"""
        return self._trace

    def set_alias(self, alias):
        """Set a user-defined identifier for this operator.  Used in
        optimization and transformation of plans"""
        self.alias = alias

    @abstractmethod
    def shortStr(self):
        """Returns a short string describing the current operator and its
        arguments, but not its children. Consider:

           query = "A(x) :- R(x,3)."
           logicalplan = dlog.fromDatalog(query)
           (label, root_op) = logicalplan[0]

           str(root_op) returns "Project($0)[Select($1 = 3)[Scan(R)]]"

           shortStr(root_op) should return "Project($0)" """

    def collectGraph(self, graph=None):
        """Collects the operator graph for a given query. Input parameter graph
        has the format {'nodes': list(), 'edges': list()}, initialized to empty
        lists by default. An input graph will be mutated."""

        # Initialize graph if necessary
        if graph is None:
            graph = {'nodes': list(), 'edges': list()}

        # Cycle detection - continue, but don't re-add this node to the graph
        if id(self) in [id(n) for n in graph['nodes']]:
            return graph

        # Add this node to the graph
        graph['nodes'].append(self)
        # Add all edges
        graph['edges'].extend([(x, self) for x in self.children()])
        for x in self.children():
            # Recursively add children and edges to the graph. This mutates
            # graph
            x.collectGraph(graph)

        # Return the graph
        return graph

    def resolveAttribute(self, ref):
        """Return a tuple of (column_name, type) for a given AttributeRef."""
        assert isinstance(ref, expression.AttributeRef), ref
        return self.scheme().resolve(ref)


class ZeroaryOperator(Operator):
    """Operator with no arguments"""
    def __init__(self):
        Operator.__init__(self)

    def __eq__(self, other):
        return self.__class__ == other.__class__

    def children(self):
        return []

    def apply(self, f):
        """Apply a function to your children"""
        return self

    def copy(self, other):
        """Deep copy"""
        Operator.copy(self, other)

    def compileme(self, resultsym):
        """Compile this operator, storing its result in resultsym"""
        raise NotImplementedError("{op}.compileme".format(op=type(self)))

    def __repr__(self):
        return "{op}()".format(op=self.opname())


class UnaryOperator(Operator):
    """Operator with one argument"""
    def __init__(self, input):
        self.input = input
        Operator.__init__(self)

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.input == other.input

    def children(self):
        return [self.input]

    def scheme(self):
        """Default scheme is the same as the input.  Usually overriden"""
        return self.input.scheme()

    def apply(self, f):
        """Apply a function to your children"""
        self.input = f(self.input)
        return self

    def copy(self, other):
        """deep copy"""
        self.input = other.input
        Operator.copy(self, other)

    def compileme(self, inputsym):
        """Compile this operator with specified input and output symbol
        names"""
        raise NotImplementedError("{op}.compileme".format(op=type(self)))

    def __repr__(self):
        return "{op}({inp!r})".format(op=self.opname(), inp=self.input)


class BinaryOperator(Operator):
    """Operator with two arguments"""
    def __init__(self, left, right):
        self.left = left
        self.right = right
        Operator.__init__(self)

    def __eq__(self, other):
        return (self.__class__ == other.__class__
                and self.left == other.left
                and self.right == other.right)

    def children(self):
        return [self.left, self.right]

    def apply(self, f):
        """Apply a function to your children"""
        self.left = f(self.left)
        self.right = f(self.right)
        return self

    def copy(self, other):
        """deep copy"""
        self.left = other.left
        self.right = other.right
        Operator.copy(self, other)

    def compileme(self, leftsym, rightsym):
        """Compile this operator with specified left, right, and output symbol
        names"""
        raise NotImplementedError("{op}.compileme".format(op=type(self)))

    def __repr__(self):
        return "{op}({l!r}, {r!r})".format(op=self.opname(), l=self.left,
                                           r=self.right)


class NaryOperator(Operator):
    """Operator with N arguments.  e.g., multi-way joins in one step."""
    def __init__(self, args=None):
        Operator.__init__(self)

        if args is None:
            self.args = []
        else:
            self.args = args

    def add(self, op):
        """Add a child operator to the end of the child argument list."""
        self.args.append(op)

    def children(self):
        return self.args

    def copy(self, other):
        """deep copy"""
        self.args = [a for a in other.args]
        Operator.copy(self, other)

    def apply(self, f):
        """Apply a function to your children"""
        self.args = [f(arg) for arg in self.args]
        return self

    def compileme(self, resultsym, argsyms):
        """Compile this operator with specified children and output symbol
        names"""
        raise NotImplementedError("{op}.compileme".format(op=type(self)))

    def __repr__(self):
        return "{op}({ch!r})".format(op=self.opname(), ch=self.args)


class NaryJoin(NaryOperator):
    """Logical Nary Join operator"""
    def __init__(self, children=None, conditions=None, output_columns=None):
        # TODO: conditions is not actually an expression, it's a list of
        # pairs of UnnamedAttributeRefs that represent equijoins. This is
        # wrong -- it should be a single expression like in Join.
        #
        # Should be able to:
        #    assert isinstance(condition, racoExpression).
        self.conditions = conditions
        self.output_columns = output_columns
        NaryOperator.__init__(self, children)

    def __eq__(self, other):
        return (NaryOperator.__eq__(self, other)
                and self.conditions == other.conditions)

    def num_tuples(self):
        # TODO: use AGM bound (P10 in http://arxiv.org/pdf/1310.3314v2.pdf)
        return DEFAULT_CARDINALITY

    def scheme(self):
        combined = reduce(operator.add, [c.scheme() for c in self.children()])
        # do projection
        if self.output_columns:
            combined = [combined[attr.get_position(combined)]
                        for attr in self.output_columns]
        return scheme.Scheme(combined)

    def copy(self, other):
        """deep copy"""
        self.conditions = other.conditions
        self.output_columns = other.output_columns
        NaryOperator.copy(self, other)

    def shortStr(self):
        return "%s(%s)" % (self.opname(), real_str(self.conditions,
                                                   skip_out=True))


"""Logical Relational Algebra"""


class Union(BinaryOperator):
    """Set union."""
    def __init__(self, left=None, right=None):
        BinaryOperator.__init__(self, left, right)

    def num_tuples(self):
        # a heuristic
        return int((self.left.num_tuples() + self.right.num_tuples()) / 2)

    def scheme(self):
        """Same semantics as SQL: Assume first schema "wins" and throw an
        error if they don't match during evaluation"""
        return self.left.scheme()

    def shortStr(self):
        return self.opname()


class UnionAll(BinaryOperator):
    """Bag union."""
    def __init__(self, left=None, right=None):
        BinaryOperator.__init__(self, left, right)

    def num_tuples(self):
        return self.left.num_tuples() + self.right.num_tuples()

    def copy(self, other):
        """deep copy"""
        BinaryOperator.copy(self, other)

    def scheme(self):
        return self.left.scheme()

    def shortStr(self):
        return self.opname()


class Intersection(BinaryOperator):
    """Set intersection."""

    def __init__(self, left=None, right=None):
        BinaryOperator.__init__(self, left, right)

    def num_tuples(self):
        return min(self.left.num_tuples(), self.right.num_tuples())

    def scheme(self):
        return self.left.scheme()

    def shortStr(self):
        return self.opname()


class Difference(BinaryOperator):
    """Set difference"""

    def __init__(self, left=None, right=None):
        BinaryOperator.__init__(self, left, right)

    def num_tuples(self):
        left_num = self.left.num_tuples()
        right_num = self.right.num_tuples()
        return left_num - math.floor(min(right_num, left_num * 0.5))

    def scheme(self):
        return self.left.scheme()

    def shortStr(self):
        return self.opname()


class CompositeBinaryOperator(BinaryOperator):
    """Join-like operations whose output schema combines its input schemas."""

    @abstractmethod
    def add_equijoin_condition(self, col0, col1):
        """Attempt to add a selection filter to this operation.

        Returns a (possibly modified) operator or None if the columns do not
        refer to different children of the join/cross-product.
        """

    @staticmethod
    def get_equijoin_condition(col0, col1):
        """Return a boolean expression representing an equijoin."""

        return expression.EQ(expression.UnnamedAttributeRef(col0),
                             expression.UnnamedAttributeRef(col1))


class CrossProduct(CompositeBinaryOperator):
    """Logical Cross Product operator"""
    def __init__(self, left=None, right=None):
        BinaryOperator.__init__(self, left, right)

    def num_tuples(self):
        return self.left.num_tuples() * self.right.num_tuples()

    def copy(self, other):
        """deep copy"""
        BinaryOperator.copy(self, other)

    def shortStr(self):
        return self.opname()

    def scheme(self):
        """Return the scheme of the result."""
        return self.left.scheme() + self.right.scheme()

    def add_equijoin_condition(self, col0, col1):
        """Convert the cross-product into a join whenever possible."""
        condition = self.get_equijoin_condition(col0, col1)
        return Join(condition, self.left, self.right)


class Join(CompositeBinaryOperator):
    """Logical Join operator"""
    def __init__(self, condition=None, left=None, right=None):
        self.condition = condition
        BinaryOperator.__init__(self, left, right)

    def __eq__(self, other):
        return (BinaryOperator.__eq__(self, other)
                and self.condition == other.condition)

    def num_tuples(self):
        # this is black magic
        return int(self.left.num_tuples() * self.right.num_tuples() / 10)

    def copy(self, other):
        """deep copy"""
        self.condition = other.condition
        BinaryOperator.copy(self, other)

    def shortStr(self):
        return "%s(%s)" % (self.opname(), self.condition)

    def scheme(self):
        """Return the scheme of the result."""
        return self.left.scheme() + self.right.scheme()

    def add_equijoin_condition(self, col0, col1):
        condition = self.get_equijoin_condition(col0, col1)
        self.condition = expression.AND(self.condition, condition)
        return self

    def __repr__(self):
        return "{op}({cond!r}, {l!r}, {r!r})".format(op=self.opname(),
                                                     cond=self.condition,
                                                     l=self.left,
                                                     r=self.right)


def resolve_attribute_name(user_name, scheme, sexpr, index):
    """Resolve an attribute/column into a name.

    :param user_name: A user-provided string name.  Can be None.
    :type user_name: string
    :param scheme: The schema of the operator's input.
    :type scheme: raco.Scheme
    :param sexpr: The scalar expression describing the column's contents.
    :type sexpr: raco.expression.Expression
    :param index: The numeric index of the column
    :type index: int
    :returns: A string representing the column name
    """

    # We always give preference to a user-provided name
    if user_name:
        return user_name

    # If the column contains a simple attribute reference, infer the column
    # name from the input schema.  However, do not pass along an auto-gen
    # column name.
    elif isinstance(sexpr, expression.AttributeRef):
        inferred_name = scheme.resolve(sexpr)[0]
        if not inferred_name.startswith('_COLUMN'):
            return inferred_name

    # Otherwise, just concoct a column name based on the column index.
    return '_COLUMN%d_' % index


class Apply(UnaryOperator):
    def __init__(self, emitters=None, input=None):
        """Create new attributes from expressions with optional rename.

        :param emitters: list of tuples of the form:
            (column_name, raco.expression.Expression).
            column_name can be None, in which case the system will infer a
            name based on the expression
        :type emitters: list of tuples
        """

        if emitters is not None:
            in_scheme = input.scheme()
            self.emitters = \
                [(resolve_attribute_name(name, in_scheme, sexpr, index), sexpr)
                 for index, (name, sexpr) in enumerate(emitters)]
        UnaryOperator.__init__(self, input)

    def __eq__(self, other):
        return (UnaryOperator.__eq__(self, other) and
                self.emitters == other.emitters)

    def num_tuples(self):
        return self.input.num_tuples()

    def copy(self, other):
        """deep copy"""
        self.emitters = other.emitters
        UnaryOperator.copy(self, other)

    def scheme(self):
        """scheme of the result."""
        input_scheme = self.input.scheme()
        new_attrs = [(name, expr.typeof(input_scheme, None))
                     for (name, expr) in self.emitters]
        return scheme.Scheme(new_attrs)

    def shortStr(self):
        estrs = ",".join(["%s=%s" % (name, str(ex))
                          for name, ex in self.emitters])
        return "%s(%s)" % (self.opname(), estrs)

    def get_names(self):
        """Get the names of the columns emitted by this Apply."""
        return [e[0] for e in self.emitters]

    def get_unnamed_emit_exprs(self):
        """Get the emit expressions for this Apply after ensuring that all
        attribute references are UnnamedAttributeRefs."""
        emits = [e[1] for e in self.emitters]
        return expression.ensure_unnamed(emits, self.input)

    def __repr__(self):
        return "{op}({emt!r}, {inp!r})".format(op=self.opname(),
                                               emt=self.emitters,
                                               inp=self.input)


class StatefulApply(UnaryOperator):
    inits = None
    updaters = None
    state_scheme = None
    emitters = None

    def __init__(self, emitters=None, state_modifiers=None, input=None):
        """Create new attributes from expressions with additional state passed
        from tuple to tuple.

        :param emitters: list of tuples of the form:
                (column_name, raco.expression.Expression).
                column_name can be None, in which case the system will infer a
                name based on the expression
            :type emitters: list of tuples
        :param state_modifiers: State variables maintained by the StatefulApply
        operator.
        :type state_modifiers: list of StateVar tuples
        """

        if state_modifiers is not None:
            self.inits = [(x.name, x.init_expr) for x in state_modifiers]
            self.updaters = [(x.name, x.update_expr) for x in state_modifiers]

            self.state_scheme = scheme.Scheme()
            for (name, expr) in self.inits:
                self.state_scheme.addAttribute(name, expr.typeof(None, None))

        if emitters is not None:
            in_scheme = input.scheme()
            self.emitters = \
                [(resolve_attribute_name(name, in_scheme, sexpr, index), sexpr)
                 for index, (name, sexpr) in enumerate(emitters)]

        UnaryOperator.__init__(self, input)

    def __eq__(self, other):
        return (super(StatefulApply, self).__eq__(self, other) and
                self.emitters == other.emitters and
                self.updaters == other.updaters and
                self.inits == other.inits)

    def __repr__(self):
        # the next line is because of the refactoring that we do in __init__
        state_mods = [StateVar(a, b, d)
                      for ((a, b), (c, d)) in zip(self.inits, self.updaters)]
        return "{op}({emt!r}, {sm!r}, {inp!r})".format(op=self.opname(),
                                                       emt=self.emitters,
                                                       sm=state_mods,
                                                       inp=self.input)

    def num_tuples(self):
        return self.input.num_tuples()

    def copy(self, other):
        """deep copy"""
        self.emitters = other.emitters
        self.updaters = other.updaters
        self.inits = other.inits
        self.state_scheme = other.state_scheme
        UnaryOperator.copy(self, other)

    def scheme(self):
        """scheme of the result."""
        input_scheme = self.input.scheme()
        new_attrs = [(name, expr.typeof(input_scheme, self.state_scheme))
                     for (name, expr) in self.emitters]
        return scheme.Scheme(new_attrs)

    def shortStr(self):
        estrs = ",".join(["%s=%s" % (name, str(ex))
                          for name, ex in self.emitters])
        return "%s(%s)" % (self.opname(), estrs)


# TODO: Non-scheme-mutating operators
class Distinct(UnaryOperator):
    """Remove duplicates from the child operator"""
    def __init__(self, input=None):
        UnaryOperator.__init__(self, input)

    def num_tuples(self):
        # TODO: better heuristics?
        return self.input.num_tuples()

    def scheme(self):
        """scheme of the result"""
        return self.input.scheme()

    def shortStr(self):
        return self.opname()


class Limit(UnaryOperator):
    def __init__(self, count=None, input=None):
        UnaryOperator.__init__(self, input)
        self.count = count

    def __eq__(self, other):
        return UnaryOperator.__eq__(self, other) and self.count == other.count

    def __repr__(self):
        return "{op}({cnt!r}, {inp!r})".format(op=self.opname(),
                                               cnt=self.count,
                                               inp=self.input)

    def num_tuples(self):
        return self.count

    def copy(self, other):
        self.count = other.count
        UnaryOperator.copy(self, other)

    def scheme(self):
        return self.input.scheme()

    def shortStr(self):
        return "%s(%s)" % (self.opname(), self.count)


class Select(UnaryOperator):
    """Logical selection operator"""
    def __init__(self, condition=None, input=None):
        self.condition = condition
        UnaryOperator.__init__(self, input)

    def __eq__(self, other):
        return (UnaryOperator.__eq__(self, other)
                and self.condition == other.condition)

    def num_tuples(self):
        return int(self.input.num_tuples() * 0.5)

    def shortStr(self):
        if isinstance(self.condition, dict):
            cond = self.condition["condition"]
        else:
            cond = self.condition
        return "%s(%s)" % (self.opname(), cond)

    def __repr__(self):
        return "{op}({cond!r}, {inp!r})".format(op=self.opname(),
                                                cond=self.condition,
                                                inp=self.input)

    def copy(self, other):
        """deep copy"""
        self.condition = other.condition
        UnaryOperator.copy(self, other)

    def scheme(self):
        """scheme of the result."""
        return self.input.scheme()

    def get_unnamed_condition(self):
        """Get the filter condition for this Select after ensuring that all
        attribute references are UnnamedAttributeRefs."""
        return expression.ensure_unnamed(self.condition, self.input)


class Project(UnaryOperator):
    """Logical projection operator"""
    def __init__(self, columnlist=None, input=None):
        self.columnlist = columnlist
        UnaryOperator.__init__(self, input)

    def __eq__(self, other):
        return (UnaryOperator.__eq__(self, other)
                and self.columnlist == other.columnlist)

    def num_tuples(self):
        return self.input.num_tuples()

    def shortStr(self):
        return "%s(%s)" % (self.opname(), real_str(self.columnlist,
                                                   skip_out=True))

    def __repr__(self):
        return "{op}({col!r}, {inp!r})".format(op=self.opname(),
                                               col=self.columnlist,
                                               inp=self.input)

    def copy(self, other):
        """deep copy"""
        self.columnlist = other.columnlist
        UnaryOperator.copy(self, other)

    def scheme(self):
        """scheme of the result. Raises a TypeError if a name in the project
        list is not in the source schema"""
        attrs = [self.input.resolveAttribute(attref)
                 for attref in self.columnlist]
        return scheme.Scheme(attrs)

    def get_unnamed_column_list(self):
        """Get the column list for this Project after ensuring that all
        attribute references are UnnamedAttributeRefs."""
        return expression.ensure_unnamed(self.columnlist, self.input)


class GroupBy(UnaryOperator):
    """Logical GroupBy operator

    :param grouping_list: A list of expressions in a "group by" clause
    :param aggregate_list: A list of aggregate expressions (e.g., MIN, MAX)
    :param input: The input operator
    :param state_modifiers: A list of StateVar tuples associated with the
    user-defined aggregates.
    """

    def __init__(self, grouping_list=None, aggregate_list=None, input=None,
                 state_modifiers=None):
        self.grouping_list = grouping_list or []
        self.aggregate_list = aggregate_list or []

        if state_modifiers is not None:
            self.inits = [(x.name, x.init_expr) for x in state_modifiers]
            self.updaters = [(x.name, x.update_expr) for x in state_modifiers]

            self.state_scheme = scheme.Scheme()
            for name, expr in self.inits:
                self.state_scheme.addAttribute(name, expr.typeof(None, None))
        else:
            self.inits = []
            self.updaters = []
            self.state_scheme = scheme.Scheme()

        UnaryOperator.__init__(self, input)

    def num_tuples(self):
        if not self.grouping_list:
            return 1
        return self.input.num_tuples()

    def shortStr(self):
        return "%s(%s; %s)" % (self.opname(),
                               real_str(self.grouping_list, skip_out=True),
                               real_str(self.aggregate_list, skip_out=True))

    def __repr__(self):
        # the next line is because of the refactoring that we do in __init__
        state_mods = [StateVar(a, b, d)
                      for ((a, b), (c, d)) in zip(self.inits, self.updaters)]

        return "{op}({gl!r}, {al!r}, {inp!r}, {sm!r})".format(
            op=self.opname(), gl=self.grouping_list, al=self.aggregate_list,
            inp=self.input, sm=state_mods)

    def copy(self, other):
        """deep copy"""
        self.grouping_list = other.grouping_list
        self.aggregate_list = other.aggregate_list
        self.updaters = other.updaters
        self.inits = other.inits
        self.state_scheme = other.state_scheme

        UnaryOperator.copy(self, other)

    def column_list(self):
        return self.grouping_list + self.aggregate_list

    def get_unnamed_grouping_list(self):
        """Get the grouping list for this GroupBy after ensuring that all
        attribute references are UnnamedAttributeRefs."""
        return expression.ensure_unnamed(self.grouping_list, self.input)

    def get_unnamed_aggregate_list(self):
        """Get the aggregate list for this GroupBy after ensuring that all
        attribute references are UnnamedAttributeRefs."""
        return expression.ensure_unnamed(self.aggregate_list, self.input)

    def get_unnamed_update_exprs(self):
        """Get the update list for this GroupBy after ensuring that all
        attribute references are UnnamedAttributeRefs."""
        ups = [expr for _, expr in self.updaters]
        return expression.ensure_unnamed(ups, self.input)

    def scheme(self):
        """scheme of the result."""
        in_scheme = self.input.scheme()
        # Note: user-provided column names are supplied by a subsequent Apply
        # invocation; see raco/myrial/groupby.py
        schema = scheme.Scheme()
        for index, sexpr in enumerate(self.column_list()):
            name = resolve_attribute_name(None, in_scheme, sexpr, index)
            _type = sexpr.typeof(in_scheme, self.state_scheme)
            schema.addAttribute(name, _type)
        return schema


class OrderBy(UnaryOperator):
    """ Logical Sort operator
    """
    def __init__(self, input=None, sort_columns=None, ascending=None):
        UnaryOperator.__init__(self, input)
        self.sort_columns = sort_columns
        self.ascending = ascending

    def num_tuples(self):
        return self.input.num_tuples()

    def shortStr(self):
        ascend_string = ['+' if a else '-' for a in self.ascending]
        sort_string = ','.join('{col}{asc}'.format(col=c, asc=a)
                               for c, a in zip(self.sort_columns,
                                               ascend_string))
        return "%s(%s)" % (self.opname(), sort_string)

    def copy(self, other):
        """deep copy"""
        self.sort_columns = other.sort_columns
        self.ascending = other.ascending
        UnaryOperator.copy(self, other)

    def scheme(self):
        return self.input.scheme()


class ProjectingJoin(Join):
    """Logical Projecting Join operator"""
    def __init__(self, condition=None, left=None, right=None,
                 output_columns=None):
        self.output_columns = output_columns
        Join.__init__(self, condition, left, right)

    def __eq__(self, other):
        return (Join.__eq__(self, other)
                and self.output_columns == other.output_columns)

    def shortStr(self):
        if self.output_columns is None:
            return Join.shortStr(self)
        return "%s(%s; %s)" % (self.opname(), self.condition,
                               real_str(self.output_columns, skip_out=True))

    def __repr__(self):
        return "{op}({cond!r}, {l!r}, {r!r}, {oc!r})"\
            .format(op=self.opname(), cond=self.condition,
                    l=self.left, r=self.right, oc=self.output_columns)

    def copy(self, other):
        """deep copy"""
        self.output_columns = other.output_columns
        Join.copy(self, other)

    def scheme(self):
        """Return the scheme of the result."""
        if self.output_columns is None:
            return Join.scheme(self)

        def get_col(pos, left_sch, right_sch):
            if pos < len(left_sch):
                return left_sch.getName(pos), left_sch.getType(pos)
            else:
                pos -= len(left_sch)
                assert pos < len(right_sch)
                return right_sch.getName(pos), right_sch.getType(pos)

        left_sch = self.left.scheme()
        right_sch = self.right.scheme()

        combined = left_sch + right_sch
        return scheme.Scheme([get_col(p.get_position(combined),
                              left_sch, right_sch)
                              for p in self.output_columns])

    def add_equijoin_condition(self, col0, col1):
        # projects are pushed after selections
        raise NotImplementedError()


class Shuffle(UnaryOperator):
    """Send the input to the specified servers"""
    def __init__(self, child=None, columnlist=None):
        UnaryOperator.__init__(self, child)
        self.columnlist = columnlist

    def num_tuples(self):
        return self.input.num_tuples()

    def shortStr(self):
        return "%s(%s)" % (self.opname(), real_str(self.columnlist,
                                                   skip_out=True))

    def copy(self, other):
        self.columnlist = other.columnlist
        UnaryOperator.copy(self, other)


class HyperCubeShuffle(UnaryOperator):
    """HyperCube Shuffle for multiway join"""
    def __init__(self, child=None, hashed_columns=None,
                 mapped_hc_dims=None, hyper_cube_dims=None,
                 cell_partition=None):
        """ Keyword arguments:
            child -- child operator.
            hashed_columns -- list of columns to be hashed.
            mapped_hc_dims --  mapped dimensions in HC of hashed columns.
            hyper_cube_dims -- size of dimensions of hyper cube.
            cell_partition -- partition of the HC cells for this shuffle.
        """
        UnaryOperator.__init__(self, child)
        self.hashed_columns = hashed_columns
        self.mapped_hc_dimensions = mapped_hc_dims
        self.hyper_cube_dimensions = hyper_cube_dims
        self.cell_partition = cell_partition

    def num_tuples(self):
        return self.input.num_tuples()

    def shortStr(self):
        return "%s(%s)" % (self.opname(), real_str(self.hashed_columns,
                                                   skip_out=True))

    def copy(self, other):
        self.hashed_columns = other.hashed_columns
        self.mapped_hc_dimensions = other.mapped_hc_dimensions
        self.hyper_cube_dimensions = other.hyper_cube_dimensions
        self.cell_partition = other.cell_partition
        UnaryOperator.copy(self, other)


class Collect(UnaryOperator):
    """Send input to one server"""
    def __init__(self, child=None, server=None):
        UnaryOperator.__init__(self, child)
        self.server = server

    def num_tuples(self):
        return self.input.num_tuples()

    def shortStr(self):
        return "%s(@%s)" % (self.opname(), self.server)

    def copy(self, other):
        self.server = other.server
        UnaryOperator.copy(self, other)


class Broadcast(UnaryOperator):
    """Send input to all servers"""
    def num_tuples(self):
        return self.input.num_tuples()

    def shortStr(self):
        return self.opname()


class PartitionBy(UnaryOperator):
    """Send input to a server indicated by a hash of specified columns."""
    def __init__(self, columnlist=None, input=None):
        self.columnlist = columnlist
        UnaryOperator.__init__(self, input)

    def __eq__(self, other):
        return (UnaryOperator.__eq__(self, other)
                and self.columnlist == other.columnlist)

    def num_tuples(self):
        return self.input.num_tuples()

    def shortStr(self):
        return "%s(%s)" % (self.opname(), real_str(self.columnlist,
                                                   skip_out=True))

    def copy(self, other):
        """deep copy"""
        self.columnlist = other.columnlist
        UnaryOperator.copy(self, other)

    def scheme(self):
        """scheme of the result. Raises a TypeError if a name in the project
        list is not in the source schema"""
        return self.input.scheme()


class Fixpoint(Operator):
    def __init__(self, body=None):
        self.body = body

    def children(self):
        return [self.body]

    def num_tuples(self):
        raise NotImplementedError("{op}.num_tuples".format(op=type(self)))

    def __str__(self):
        return "%s[%s]" % (self.shortStr(), self.body)

    def shortStr(self):
        return """Fixpoint"""

    def apply(self, f):
        """Apply a function to your children"""
        self.body.apply(f)
        return self

    def scheme(self):
        if self.body:
            return self.body.scheme()
        else:
            raise RecursionError("No Scheme defined yet for fixpoint")

    def loopBody(self, plan):
        self.body = plan


class State(ZeroaryOperator):
    """A placeholder operator for a recursive plan"""

    def __init__(self, name, fixpoint):
        ZeroaryOperator.__init__(self)
        self.name = name
        self.fixpoint = fixpoint

    def shortStr(self):
        return "%s(%s)" % (self.opname(), self.name)

    def scheme(self):
        return self.fixpoint.scheme()


class Store(UnaryOperator):
    """Store output to a relational table.

    relation_key is a string of the form "program:user:relation".
    """

    def __init__(self, relation_key=None, plan=None):
        UnaryOperator.__init__(self, plan)
        self.relation_key = relation_key

    def num_tuples(self):
        return self.input.num_tuples()

    def shortStr(self):
        return "%s(%s)" % (self.opname(), self.relation_key)

    def __repr__(self):
        return "{op}({rk!r}, {pl!r})".format(op=self.opname(),
                                             rk=self.relation_key,
                                             pl=self.input)

    def copy(self, other):
        self.relation_key = other.relation_key
        UnaryOperator.copy(self, other)


class Dump(UnaryOperator):
    """Echo input to standard out; only useful for standalone raco."""

    def num_tuples(self):
        raise NotImplementedError("{op}.num_tuples".format(op=type(self)))

    def shortStr(self):
        return "%s()" % self.opname()


class EmptyRelation(ZeroaryOperator):
    """Relation with no tuples."""

    def __init__(self, _scheme=None):
        self._scheme = _scheme

    def num_tuples(self):
        return 0

    def shortStr(self):
        return "%s(%s)" % (self.opname(), self._scheme)

    def __repr__(self):
        return "{op}({sch!r})".format(op=self.opname(), sch=self._scheme)

    def copy(self, other):
        """deep copy"""
        self._scheme = other._scheme

    def scheme(self):
        """scheme of the result."""
        return self._scheme


class SingletonRelation(ZeroaryOperator):
    """Relation with a single empty tuple.

    Used for constructing table literals.
    """

    def shortStr(self):
        return "SingletonRelation"

    def num_tuples(self):
        return 1

    def copy(self, other):
        """deep copy"""
        pass

    def scheme(self):
        """scheme of the result."""
        return scheme.Scheme()


class FileScan(ZeroaryOperator):
    """Load table data from a file."""

    def __init__(self, path=None, _scheme=None):
        self.path = path
        self._scheme = _scheme
        ZeroaryOperator.__init__(self)

    def __eq__(self, other):
        return (ZeroaryOperator.__eq__(self, other)
                and self.path == other.path
                and self.scheme() == other.scheme())

    def __hash__(self):
        return ("%s-%s" % (self.opname(), self.path)).__hash__()

    def shortStr(self):
        return "%s(%s)" % (self.opname(), self.path)

    def __repr__(self):
        return "{op}({path!r}, {sch!r})".format(op=self.opname(),
                                                path=self.path,
                                                sch=self._scheme)

    def num_tuples(self):
        raise NotImplementedError("{op}.num_tuples".format(op=type(self)))

    def copy(self, other):
        """deep copy"""
        self.path = other.path
        self._scheme = other._scheme

        ZeroaryOperator.copy(self, other)

    def scheme(self):
        return self._scheme


class Scan(ZeroaryOperator):
    """Logical Scan operator."""

    def __init__(self, relation_key=None, _scheme=None, cardinality=None):
        """Initialize a scan operator.

        relation_key is a string of the form "user:program:relation"
        scheme is the schema of the relation.
        """
        self.relation_key = relation_key
        self._scheme = _scheme
        if cardinality is not None:
            self._cardinality = cardinality
        else:
            self._cardinality = DEFAULT_CARDINALITY
        ZeroaryOperator.__init__(self)

    def __eq__(self, other):
        return (ZeroaryOperator.__eq__(self, other)
                and self.relation_key == other.relation_key
                and self.scheme() == other.scheme())

    def __hash__(self):
        """
        Override since Scan is Zeroary and needs other distinguishing aspects
        to avoid collisions
        """
        return ("%s-%s" % (self.opname(), self.relation_key)).__hash__()

    def shortStr(self):
        return "%s(%s)" % (self.opname(), self.relation_key)

    def num_tuples(self):
        return self._cardinality

    def __repr__(self):
        return "{op}({rk!r}, {sch!r}, {card!r})".format(
            op=self.opname(), rk=self.relation_key, sch=self._scheme,
            card=self._cardinality)

    def copy(self, other):
        """deep copy"""
        self.relation_key = other.relation_key
        self._scheme = other._scheme
        self._cardinality = other._cardinality

        # TODO: need a cleaner and more general way of tracing information
        # through the compilation process for debugging purposes
        if hasattr(other, "originalterm"):
            self.originalterm = other.originalterm
        ZeroaryOperator.copy(self, other)

    def scheme(self):
        """Scheme of the result, which is just the scheme of the relation."""
        return self._scheme


class StoreTemp(UnaryOperator):
    """Store an input relation to a "temporary" relation.

    Temporary relations exist for the lifetime of a query.
    """
    def __init__(self, name=None, input=None):
        UnaryOperator.__init__(self, input)
        self.name = name

    def num_tuples(self):
        return self.input.num_tuples()

    def shortStr(self):
        return '{op}({name})'.format(op=self.opname(), name=self.name)

    def copy(self, other):
        self.name = other.name
        UnaryOperator.copy(self, other)

    def __eq__(self, other):
        return UnaryOperator.__eq__(self, other) and self.name == other.name

    def __repr__(self):
        return "{op}({name!r}, {inp!r})".format(op=self.opname(),
                                                name=self.name,
                                                inp=self.input)


class AppendTemp(UnaryOperator):
    """Append an input relation to a "temporary" relation.

    Temporary relations exist for the lifetime of a query.
    """
    def __init__(self, name=None, input=None):
        UnaryOperator.__init__(self, input)
        self.name = name

    def num_tuples(self):
        return self.input.num_tuples()

    def shortStr(self):
        return '{op}({name})'.format(op=self.opname(), name=self.name)

    def copy(self, other):
        self.name = other.name
        UnaryOperator.copy(self, other)

    def __eq__(self, other):
        return UnaryOperator.__eq__(self, other) and self.name == other.name

    def __repr__(self):
        return "{op}({name!r}, {inp!r})".format(op=self.opname(),
                                                name=self.name,
                                                inp=self.input)


class ScanTemp(ZeroaryOperator):
    """Read the contents of a temporary relation."""

    def __init__(self, name=None, _scheme=None):
        self.name = name
        self._scheme = _scheme
        ZeroaryOperator.__init__(self)

    def __eq__(self, other):
        return (ZeroaryOperator.__eq__(self, other) and self.name == other.name
                and self._scheme == other._scheme)

    def num_tuples(self):
        raise NotImplementedError("{op}.num_tuples".format(op=type(self)))

    def shortStr(self):
        return "%s(%s,%s)" % (self.opname(), self.name, str(self._scheme))

    def copy(self, other):
        self.name = other.name
        self._scheme = other._scheme
        ZeroaryOperator.copy(self, other)

    def scheme(self):
        return self._scheme

    def __repr__(self):
        return "{op}({name!r}, {sch!r})".format(op=self.opname(),
                                                name=self.name,
                                                sch=self._scheme)


class Parallel(NaryOperator):
    """Execute a set of independent plans in parallel."""
    def __init__(self, ops=None):
        NaryOperator.__init__(self, ops)

    def num_tuples(self):
        raise NotImplementedError("{op}.num_tuples".format(op=type(self)))

    def shortStr(self):
        return self.opname()

    def scheme(self):
        """Parallel does not return any tuples."""
        return None


class Sequence(NaryOperator):
    """Execute a sequence of plans in serial order."""
    def __init__(self, ops=None):
        NaryOperator.__init__(self, ops)

    def shortStr(self):
        return self.opname()

    def scheme(self):
        """Sequence does not return any tuples."""
        return None

    def __repr__(self):
        return "{op}({ops!r})".format(op=self.opname(), ops=self.args)

    def num_tuples(self):
        raise NotImplementedError("{op}.num_tuples".format(op=type(self)))


class DoWhile(NaryOperator):
    def __init__(self, ops=None):
        """Repeatedly execute a sequence of plans until a termination
        condition.

        :params ops: A list of operations to execute in serial.  By convention,
        the last operation is the termination condition.  The termination
        condition should map to a single row, single column relation.  The loop
        continues if its value is True.
        """
        if ops is not None:
            assert len(ops) >= 2, "DoWhile should have at least two children"
        NaryOperator.__init__(self, ops)

    def num_tuples(self):
        raise NotImplementedError("{op}.num_tuples".format(op=type(self)))

    def shortStr(self):
        return self.opname()

    def scheme(self):
        """DoWhile does not return any tuples."""
        return None


def inline_operator(dest_op, var, target_op):
    """Convert two operator trees into one by inlining.

    :param dest_op: The Operator that is the inline destination
    :param var: The variable name (String) to replace.
    :param target_op: The operation to replace.
    """
    # Wrap the bool in a list so we pass a pointer that does not change
    # (the list) into the function.
    has_inlined = [False]

    def rewrite_node(node):
        if isinstance(node, ScanTemp) and node.name == var:
            if has_inlined[0]:
                return copy.deepcopy(target_op)
            has_inlined[0] = True
            return target_op
        else:
            return node.apply(rewrite_node)

    return rewrite_node(dest_op)




"""
Implemention of groupby/aggregation in Myrial.

The inputs are a child Operator and a list of column mappings, each defined
by tuples of the form: (column_name, scalar_expression).  The outputs are
(possibly) revised operators and list of column mappings.

The basic algorithm:

1) Scan the list of columns looking for any aggregte expressions.  If none
are found, then we return the inputs unmodified.

2) Next, we scan each column, switching on whether the scalar expression
contains any aggregate expression:

2A) Columns without aggregate expressions are "groupby" terms.  We add such
terms to a list of groupby terms.

2B) Columns with aggregate expressions are "aggregation" terms.  We record all
aggregate expressions in a list (actually, an OrderedDict).  And, we apply
a "hoisting" procedure whereby each aggregate expression is replaced by
a reference to a raw column reference.  As an example:

range = max(salary) - min(salary)
==>
range = $4 - $5

3) We create a GroupBy relational algebra operation with the grouping terms
and aggregate terms calculated in steps 2A and 2B.

4) We return an updated set of column mappings; the Myrial interpreter
uses these mappings to form an Apply operator that stitches up the column
names and values as expected by the caller.
"""

import collections
import raco.expression
from raco.myrial.exceptions import *


class NonGroupedAccessException(Exception):
    """Attempting to access a non-grouping term in an aggregate expression"""
    pass


class AggregateState(object):
    def __init__(self, initial_aggregate_pos):
        # Mapping from an aggregate scalar expression (e.g., MAX(salary)) to
        # a non-aggregate scalar expression (a raw column index).
        self.aggregates = collections.OrderedDict()

        # Next index to be assigned for aggregate expressions
        self.aggregate_pos = initial_aggregate_pos


def __hoist_aggregates(sexpr, agg_state, group_mappings, input_scheme):
    def hoist_node(sexpr):
        if isinstance(sexpr, raco.expression.AttributeRef):
            # Translate the attribute ref to the schema that will exist
            # after the GroupBy
            input_pos = sexpr.get_position(input_scheme)
            if input_pos not in group_mappings:
                raise NonGroupedAccessException(str(sexpr))
            output_pos = group_mappings[input_pos]
            return raco.expression.UnnamedAttributeRef(output_pos)

        if not isinstance(sexpr, raco.expression.AggregateExpression):
            return sexpr

        if sexpr in agg_state.aggregates:
            return agg_state.aggregates[sexpr]
        else:
            out = raco.expression.UnnamedAttributeRef(agg_state.aggregate_pos)
            agg_state.aggregates[sexpr] = out
            agg_state.aggregate_pos += 1
            return out

    def recursive_eval(sexpr):
        """Apply hoisting to a scalar expression and all its descendents"""
        newexpr = hoist_node(sexpr)
        newexpr.apply(recursive_eval)
        return newexpr

    return recursive_eval(sexpr)


def groupby(op, emit_clause, extra_grouping_columns, statemods=None):
    """Process groupby/aggregation expressions."""

    assert emit_clause

    # A mapping from input position (before the GroupBy) to output position
    # (after the GroupBy) for grouping terms.  This allows aggregate terms
    # to refer to grouping fields.
    group_mappings = {}

    scheme = op.scheme()
    num_group_terms = 0

    for name, sexpr in emit_clause:
        if not raco.expression.expression_contains_aggregate(sexpr):
            if isinstance(sexpr, raco.expression.AttributeRef):
                group_mappings[sexpr.get_position(scheme)] = num_group_terms
            num_group_terms += 1

    # The user must have specified an aggregate expression to trigger
    # a groupby invocation.
    assert num_group_terms != len(emit_clause)

    # Add extra grouping columns; we group by these terms, but the output
    # is not preserved in the final apply invocation.  These are columns
    # that were referenced in unbox expressions.
    for col in extra_grouping_columns:
        group_mappings[col] = num_group_terms
        num_group_terms += 1

    # State about scalar expressions with aggregates
    agg_state = AggregateState(num_group_terms)

    # mappings from column name to scalar expressions; these mappings are
    # applied after the GroupBy operator to stitch up column names and values.
    output_mappings = []

    # A subset of the scalar expressions in the emit clause that do
    # not contain aggregate expressions; these become the grouping terms
    # to the GroupBy operator.
    group_terms = []

    for name, sexpr in emit_clause:
        if raco.expression.expression_contains_aggregate(sexpr):
            output_mappings.append(
                (name, __hoist_aggregates(sexpr, agg_state, group_mappings,
                                          scheme)))
        else:
            output_mappings.append((
                name, raco.expression.UnnamedAttributeRef(len(group_terms))))
            group_terms.append(sexpr)

    # Add extra grouping columns; note that these are not present in the
    # output mappings.
    group_terms.extend([raco.expression.UnnamedAttributeRef(c)
                        for c in extra_grouping_columns])

    agg_terms = agg_state.aggregates.keys()
    op1 = raco.algebra.GroupBy(group_terms, agg_terms, op, statemods)
    return raco.algebra.Apply(emitters=output_mappings, input=op1)



import json
import unittest

from raco.language.myrialang import compile_to_json
import raco.fakedb
from raco.language.myrialang import MyriaLeftDeepTreeAlgebra
import raco.myrial.interpreter as interpreter
import raco.myrial.parser as parser
import raco.viz
from raco.replace_with_repr import replace_with_repr


class MyrialTestCase(unittest.TestCase):
    def create_db(self):
        return raco.fakedb.FakeDatabase()

    def setUp(self):
        self.db = self.create_db()

        self.parser = parser.Parser()
        self.processor = interpreter.StatementProcessor(self.db)

    def parse(self, query):
        """Parse a query"""
        statements = self.parser.parse(query)
        self.processor.evaluate(statements)

    def get_plan(self, query, **kwargs):
        """Get the MyriaL query plan for a query"""
        statements = self.parser.parse(query)
        self.processor.evaluate(statements)
        if kwargs.get('logical', False):
            p = self.processor.get_logical_plan(**kwargs)
        else:
            p = self.processor.get_physical_plan(**kwargs)
        # verify that we can stringify p
        # TODO verify the string somehow?
        assert str(p)
        # verify that we can convert p to a dot
        # TODO verify the dot somehow?
        raco.viz.get_dot(p)
        # Test repr
        return replace_with_repr(p)

    def get_logical_plan(self, query, **kwargs):
        """Get the logical plan for a MyriaL query"""
        kwargs['logical'] = True
        return self.get_plan(query, **kwargs)

    def get_physical_plan(self, query, **kwargs):
        """Get the physical plan for a MyriaL query"""
        kwargs['logical'] = False
        return self.get_plan(query, **kwargs)

    def execute_query(self, query, test_logical=False, skip_json=False,
                      output='OUTPUT'):
        """Run a test query against the fake database"""
        plan = self.get_plan(query, logical=test_logical)

        if not test_logical and not skip_json:
            # Test that JSON compilation runs without error
            # TODO: verify the JSON output somehow?
            json_string = json.dumps(compile_to_json(
                "some query", "some logical plan", plan, "myrial"))
            assert json_string

        self.db.evaluate(plan)

        return self.db.get_table(output)

    def check_result(self, query, expected, test_logical=False,
                     skip_json=False, output='OUTPUT', scheme=None):
        """Execute a test query with an expected output"""
        actual = self.execute_query(query, test_logical, skip_json, output)
        self.assertEquals(actual, expected)

        if scheme:
            self.assertEquals(self.db.get_scheme(output), scheme)

    def check_scheme(self, query, scheme):
        """Execute a test query with an expected output schema."""
        actual = self.execute_query(query)
        self.assertEquals(self.db.get_scheme('OUTPUT'), scheme)



from raco import algebra
from raco import expression
from expression import (accessed_columns, UnnamedAttributeRef,
                        to_unnamed_recursive)

from abc import ABCMeta, abstractmethod
import itertools


class Rule(object):
    """Argument is an expression tree

    Returns a possibly modified expression tree"""

    __metaclass__ = ABCMeta

    def __call__(self, expr):
        return self.fire(expr)

    @abstractmethod
    def fire(self, expr):
        """Apply this rule to the supplied expression tree"""


class CrossProduct2Join(Rule):
    """A rewrite rule for removing Cross Product"""
    def fire(self, expr):
        if isinstance(expr, algebra.CrossProduct):
            return algebra.Join(expression.EQ(expression.NumericLiteral(1),
                                expression.NumericLiteral(1)),
                                expr.left, expr.right)
        return expr

    def __str__(self):
        return "CrossProduct(left, right) => Join(1=1, left, right)"


class removeProject(Rule):
    """A rewrite rule for removing Projections"""
    def fire(self, expr):
        if isinstance(expr, algebra.Project):
            return expr.input
        return expr

    def __str__(self):
        return "Project => ()"


class OneToOne(Rule):
    def __init__(self, opfrom, opto):
        self.opfrom = opfrom
        self.opto = opto

    def fire(self, expr):
        if isinstance(expr, self.opfrom):
            newop = self.opto()
            newop.copy(expr)
            return newop
        return expr

    def __str__(self):
        return "%s => %s" % (self.opfrom.__name__, self.opto.__name__)


class JoinToProjectingJoin(Rule):
    """A rewrite rule for turning every Join into a ProjectingJoin"""

    def fire(self, expr):
        if not isinstance(expr, algebra.Join) or \
                isinstance(expr, algebra.ProjectingJoin):
            return expr

        return algebra.ProjectingJoin(expr.condition,
                                      expr.left, expr.right,
                                      expr.scheme().ascolumnlist())

    def __str__(self):
        return "Join => ProjectingJoin"


def is_simple_agg_expr(agg):
    """A simple aggregate expression is an aggregate whose input is an
    AttributeRef."""
    if isinstance(agg, expression.COUNTALL):
        return True
    elif isinstance(agg, expression.UdaAggregateExpression):
        return True
    elif (isinstance(agg, expression.UnaryOperator) and
          isinstance(agg, expression.BuiltinAggregateExpression) and
          isinstance(agg.input, expression.AttributeRef)):
        return True
    return False


class SimpleGroupBy(Rule):
    # A "Simple" GroupBy is one that has only AttributeRefs as its grouping
    # fields, and only AggregateExpression(AttributeRef) as its aggregate
    # fields.
    #
    # Even AggregateExpression(Literal) is more complicated than Myria's
    # GroupBy wants to handle. Thus we will insert Apply before a GroupBy to
    # take all the "Complex" expressions away.

    def fire(self, expr):
        if not isinstance(expr, algebra.GroupBy):
            return expr

        # A simple grouping expression is an AttributeRef
        def is_simple_grp_expr(grp):
            return isinstance(grp, expression.AttributeRef)

        complex_grp_exprs = [(i, grp)
                             for (i, grp) in enumerate(expr.grouping_list)
                             if not is_simple_grp_expr(grp)]

        complex_agg_exprs = [agg for agg in expr.aggregate_list
                             if not is_simple_agg_expr(agg)]

        # There are no complicated expressions, we're okay with the existing
        # GroupBy.
        if not complex_grp_exprs and not complex_agg_exprs:
            return expr

        # Construct the Apply we're going to stick before the GroupBy

        child_scheme = expr.input.scheme()

        # First: copy every column from the input verbatim
        mappings = [(None, UnnamedAttributeRef(i))
                    for i in range(len(child_scheme))]

        # Next: move the complex grouping expressions into the Apply, replace
        # with simple refs
        for i, grp_expr in complex_grp_exprs:
            mappings.append((None, grp_expr))
            expr.grouping_list[i] = UnnamedAttributeRef(len(mappings) - 1)

        # Finally: move the complex aggregate expressions into the Apply,
        # replace with simple refs
        for agg_expr in complex_agg_exprs:
            mappings.append((None, agg_expr.input))
            agg_expr.input = UnnamedAttributeRef(len(mappings) - 1)

        # Construct and prepend the new Apply
        new_apply = algebra.Apply(mappings, expr.input)
        expr.input = new_apply

        # Don't overwrite expr.grouping_list or expr.aggregate_list, instead we
        # are mutating the objects it contains when we modify grp_expr or
        # agg_expr in the above for loops.
        return expr


class DedupGroupBy(Rule):
    """When a GroupBy computes redundant fields, replace this duplicate
    computation by a single computation plus a duplicating Apply."""

    def fire(self, expr):
        if not isinstance(expr, algebra.GroupBy):
            return expr

        aggs = expr.get_unnamed_aggregate_list()

        # Maps aggregate index j to index i < j that j duplicates
        dups = {}
        # Maps non-dup aggregate index i to index i' <= i it will have
        # in the output aggregate list.
        orig = {}
        for i, a in enumerate(aggs):
            if i in dups:
                continue
            orig[i] = len(orig)
            dups.update({(j + i + 1): i
                         for j, b in enumerate(aggs[(i + 1):])
                         if a == b})

        if len(dups) == 0:
            # All the aggregate expressions are unique, we're good
            return expr

        #################################
        # Construct a new Apply that drops all duplicates and replaces
        # them with repeated UnnamedAttributeRefs
        #################################

        # First keep the grouping list intact
        num_grps = len(expr.grouping_list)
        mappings = [(None, UnnamedAttributeRef(i))
                    for i in range(num_grps)]

        # Construct the references to the grouping list
        for i in range(len(aggs)):
            if i in orig:
                m = orig[i] + num_grps
            else:
                m = orig[dups[i]] + num_grps
            mappings.append((None, UnnamedAttributeRef(m)))

        # Drop any duplicates from the agg list
        expr.aggregate_list = [aggs[i] for i in sorted(orig)]

        return algebra.Apply(emitters=mappings, input=expr)


class DistinctToGroupBy(Rule):
    """Turns a distinct into an empty GroupBy"""
    def fire(self, expr):
        if isinstance(expr, algebra.Distinct):
            in_scheme = expr.scheme()
            grps = [UnnamedAttributeRef(i) for i in range(len(in_scheme))]
            return algebra.GroupBy(input=expr.input, grouping_list=grps)

        return expr

    def __str__(self):
        return "Distinct => GroupBy"


class EmptyGroupByToDistinct(Rule):
    """Turns a GroupBy with no aggregates into a Distinct"""
    def fire(self, expr):
        if isinstance(expr, algebra.GroupBy) and len(expr.aggregate_list) == 0:
            # We can turn an empty GroupBy into a Distinct. However,
            # we must ensure that the GroupBy does not do any column
            # re-ordering.
            group_cols = expr.get_unnamed_grouping_list()
            if all(e.position == i for i, e in enumerate(group_cols)):
                # No reordering is done
                return algebra.Distinct(input=expr.input)

            # Some reordering is done, so shim in the Apply to mimic it.
            reorder_cols = algebra.Apply(
                emitters=[(None, e) for e in group_cols], input=expr.input)
            return algebra.Distinct(input=reorder_cols)

        return expr

    def __str__(self):
        return "Distinct => GroupBy"


class CountToCountall(Rule):
    """Since Raco does not support NULLs at the moment, it is safe to always
    map COUNT to COUNTALL."""
    # TODO fix when we have NULL support.
    def fire(self, expr):
        if not isinstance(expr, algebra.GroupBy):
            return expr

        assert all(is_simple_agg_expr(agg) for agg in expr.aggregate_list)

        counts = [i for (i, agg) in enumerate(expr.aggregate_list)
                  if isinstance(agg, expression.COUNT)]
        if not counts:
            return expr

        for i in counts:
            expr.aggregate_list[i] = expression.COUNTALL()

        return expr

    def __str__(self):
        return "Count(x) => Countall()"


class RemoveTrivialSequences(Rule):
    def fire(self, expr):
        if not isinstance(expr, algebra.Sequence):
            return expr

        if len(expr.args) == 1:
            return expr.args[0]
        else:
            return expr


class SplitSelects(Rule):
    """Replace AND clauses with multiple consecutive selects."""

    def fire(self, op):
        if not isinstance(op, algebra.Select):
            return op

        conjuncs = expression.extract_conjuncs(op.get_unnamed_condition())
        assert conjuncs  # Must be at least 1

        op.condition = conjuncs[0]
        op.has_been_pushed = False
        for conjunc in conjuncs[1:]:
            op = algebra.Select(conjunc, op)
            op.has_been_pushed = False
        return op

    def __str__(self):
        return "Select => Select, Select"


class PushSelects(Rule):
    """Push selections."""

    @staticmethod
    def is_column_equality_comparison(cond):
        """Return a tuple of column indexes if the condition is an equality
        test.
        """

        if (isinstance(cond, expression.EQ) and
                isinstance(cond.left, UnnamedAttributeRef) and
                isinstance(cond.right, UnnamedAttributeRef)):
            return cond.left.position, cond.right.position
        else:
            return None

    @staticmethod
    def descend_tree(op, cond):
        """Recursively push a selection condition down a tree of operators.

        :param op: The root of an operator tree
        :type op: raco.algebra.Operator
        :type cond: The selection condition
        :type cond: raco.expression.expression

        :return: A (possibly modified) operator.
        """

        if isinstance(op, algebra.Select):
            # Keep pushing; selects are commutative
            op.input = PushSelects.descend_tree(op.input, cond)
            return op
        elif isinstance(op, algebra.CompositeBinaryOperator):
            # Joins and cross-products; consider conversion to an equijoin
            left_len = len(op.left.scheme())
            accessed = accessed_columns(cond)
            in_left = [col < left_len for col in accessed]
            if all(in_left):
                # Push the select into the left sub-tree.
                op.left = PushSelects.descend_tree(op.left, cond)
                return op
            elif not any(in_left):
                # Push into right subtree; rebase column indexes
                expression.rebase_expr(cond, left_len)
                op.right = PushSelects.descend_tree(op.right, cond)
                return op
            else:
                # Selection includes both children; attempt to create an
                # equijoin condition
                cols = PushSelects.is_column_equality_comparison(cond)
                if cols:
                    return op.add_equijoin_condition(cols[0], cols[1])
        elif isinstance(op, algebra.Apply):
            # Convert accessed to a list from a set to ensure consistent order
            accessed = list(accessed_columns(cond))
            accessed_emits = [op.emitters[i][1] for i in accessed]
            if all(isinstance(e, expression.AttributeRef)
                   for e in accessed_emits):
                unnamed_emits = expression.ensure_unnamed(
                    accessed_emits, op.input)
                # This condition only touches columns that are copied verbatim
                # from the child, so we can push it.
                index_map = {a: e.position
                             for (a, e) in zip(accessed, unnamed_emits)}
                expression.reindex_expr(cond, index_map)
                op.input = PushSelects.descend_tree(op.input, cond)
                return op
        elif isinstance(op, algebra.GroupBy):
            # Convert accessed to a list from a set to ensure consistent order
            accessed = list(accessed_columns(cond))
            if all((a < len(op.grouping_list)) for a in accessed):
                accessed_grps = [op.grouping_list[a] for a in accessed]
                # This condition only touches columns that are copied verbatim
                # from the child (grouping keys), so we can push it.
                assert all(isinstance(e, expression.AttributeRef)
                           for e in op.grouping_list)
                unnamed_grps = expression.ensure_unnamed(accessed_grps,
                                                         op.input)
                index_map = {a: e.position
                             for (a, e) in zip(accessed, unnamed_grps)}
                expression.reindex_expr(cond, index_map)
                op.input = PushSelects.descend_tree(op.input, cond)
                return op

        # Can't push any more: instantiate the selection
        new_op = algebra.Select(cond, op)
        new_op.has_been_pushed = True
        return new_op

    def fire(self, op):
        if not isinstance(op, algebra.Select):
            return op
        if op.has_been_pushed:
            return op

        new_op = PushSelects.descend_tree(op.input, op.condition)

        # The new root may also be a select, so fire the rule recursively
        return self.fire(new_op)

    def __str__(self):
        return ("Select, Cross/Join => Join;"
                + " Select, Apply => Apply, Select;"
                + " Select, GroupBy => GroupBy, Select")


class MergeSelects(Rule):
    """Merge consecutive Selects into a single conjunctive selection."""

    def fire(self, op):
        if not isinstance(op, algebra.Select):
            return op

        while isinstance(op.input, algebra.Select):
            conjunc = expression.AND(op.condition, op.input.condition)
            op = algebra.Select(conjunc, op.input.input)

        return op

    def __str__(self):
        return "Select, Select => Select"


class PushApply(Rule):
    """Many Applies in MyriaL are added to select fewer columns from the
    input. In some  of these cases, we can do less work in the children by
    preventing them from producing columns we will then immediately drop.

    Currently, this rule:
      - merges consecutive Apply operations into one Apply, possibly dropping
        some of the produced columns along the way.
      - makes ProjectingJoin only produce columns that are later read.
    """

    def fire(self, op):
        if not isinstance(op, algebra.Apply):
            return op

        child = op.input

        if isinstance(child, algebra.Apply):

            emits = op.get_unnamed_emit_exprs()
            child_emits = child.get_unnamed_emit_exprs()

            def convert(n):
                if isinstance(n, expression.UnnamedAttributeRef):
                    return child_emits[n.position]
                n.apply(convert)
                return n
            emits = [convert(e) for e in emits]

            new_apply = algebra.Apply(emitters=zip(op.get_names(), emits),
                                      input=child.input)
            return self.fire(new_apply)

        elif isinstance(child, algebra.ProjectingJoin):
            emits = op.get_unnamed_emit_exprs()

            # If this apply is only AttributeRefs and the columns already
            # have the correct names, we can push it into the ProjectingJoin
            if (all(isinstance(e, expression.AttributeRef) for e in emits)
                    and len(set(emits)) == len(emits)):
                new_cols = [child.output_columns[e.position] for e in emits]
                # We need to ensure that left columns come before right cols
                left_sch = child.left.scheme()
                right_sch = child.right.scheme()
                combined = left_sch + right_sch
                left_len = len(left_sch)
                new_cols = [expression.to_unnamed_recursive(e, combined)
                            for e in new_cols]
                side = [e.position >= left_len for e in new_cols]
                if sorted(side) == side:
                    # Left columns do come before right cols
                    new_pj = algebra.ProjectingJoin(
                        condition=child.condition, left=child.left,
                        right=child.right, output_columns=new_cols)
                    if new_pj.scheme() == op.scheme():
                        return new_pj

            accessed = sorted(set(itertools.chain(*(accessed_columns(e)
                                                    for e in emits))))
            index_map = {a: i for (i, a) in enumerate(accessed)}
            child.output_columns = [child.output_columns[i] for i in accessed]
            for e in emits:
                expression.reindex_expr(e, index_map)

            return algebra.Apply(emitters=zip(op.get_names(), emits),
                                 input=child)

        elif isinstance(child, algebra.GroupBy):
            emits = op.get_unnamed_emit_exprs()
            assert all(is_simple_agg_expr(agg) for agg in child.aggregate_list)

            accessed = sorted(set(itertools.chain(*(accessed_columns(e)
                                                    for e in emits))))
            num_grps = len(child.grouping_list)
            accessed_aggs = [i for i in accessed if i >= num_grps]
            if len(accessed_aggs) == len(child.aggregate_list):
                return op

            unused_map = {i: j + num_grps for j, i in enumerate(accessed_aggs)}
            child.aggregate_list = [child.aggregate_list[i - num_grps]
                                    for i in accessed_aggs]
            for e in emits:
                expression.reindex_expr(e, unused_map)

            return algebra.Apply(emitters=zip(op.get_names(), emits),
                                 input=child)

        return op

    def __str__(self):
        return 'Push Apply into Apply, ProjectingJoin'


class ProjectToDistinctColumnSelect(Rule):
    def fire(self, expr):
        # If not a Project, who cares?
        if not isinstance(expr, algebra.Project):
            return expr

        mappings = [(None, x) for x in expr.columnlist]
        col_select = algebra.Apply(mappings, expr.input)
        return algebra.Distinct(input=col_select)

    def __str__(self):
        return 'Project => Distinct, Column select'


class RemoveUnusedColumns(Rule):
    """For operators that construct new tuples (e.g., GroupBy or Join), we are
    guaranteed that any columns from an input tuple that are ignored (neither
    used internally nor to produce the output columns) cannot be used higher
    in the query tree. For these cases, this rule will prepend an Apply that
    keeps only the referenced columns. The goal is that after this rule,
    a subsequent invocation of PushApply will be able to push that
    column-selection operation further down the tree."""

    def fire(self, op):
        if isinstance(op, algebra.GroupBy):
            child = op.input
            child_scheme = child.scheme()
            grp_list = op.get_unnamed_grouping_list()
            agg_list = op.get_unnamed_aggregate_list()

            up_names = [name for name, ex in op.updaters]
            up_list = op.get_unnamed_update_exprs()

            agg = [accessed_columns(a) for a in agg_list]
            up = [accessed_columns(a) for a in up_list]
            pos = [{g.position} for g in grp_list]

            accessed = sorted(set(itertools.chain(*(up + agg + pos))))
            if not accessed:
                # Bug #207: COUNTALL() does not access any columns. So if the
                # query is just a COUNT(*), we would generate an empty Apply.
                # If this happens, just keep the first column of the input.
                accessed = [0]
            if len(accessed) != len(child_scheme):
                emitters = [(None, UnnamedAttributeRef(i)) for i in accessed]
                new_apply = algebra.Apply(emitters, child)
                index_map = {a: i for (i, a) in enumerate(accessed)}
                for agg_expr in itertools.chain(grp_list, agg_list, up_list):
                    expression.reindex_expr(agg_expr, index_map)
                op.grouping_list = grp_list
                op.aggregate_list = agg_list
                op.updaters = [(name, ex) for name, ex in
                               zip(up_names, up_list)]
                op.input = new_apply
                return op
        elif isinstance(op, algebra.ProjectingJoin):
            l_scheme = op.left.scheme()
            r_scheme = op.right.scheme()
            in_scheme = l_scheme + r_scheme
            condition = to_unnamed_recursive(op.condition, in_scheme)
            column_list = [to_unnamed_recursive(c, in_scheme)
                           for c in op.output_columns]

            accessed = (accessed_columns(condition)
                        | set(c.position for c in op.output_columns))
            if len(accessed) == len(in_scheme):
                return op

            accessed = sorted(accessed)
            left = [a for a in accessed if a < len(l_scheme)]
            if len(left) < len(l_scheme):
                emits = [(None, UnnamedAttributeRef(a)) for a in left]
                apply = algebra.Apply(emits, op.left)
                op.left = apply
            right = [a - len(l_scheme) for a in accessed
                     if a >= len(l_scheme)]
            if len(right) < len(r_scheme):
                emits = [(None, UnnamedAttributeRef(a)) for a in right]
                apply = algebra.Apply(emits, op.right)
                op.right = apply
            index_map = {a: i for (i, a) in enumerate(accessed)}
            expression.reindex_expr(condition, index_map)
            [expression.reindex_expr(c, index_map) for c in column_list]
            op.condition = condition
            op.output_columns = column_list
            return op

        return op

    def __str__(self):
        return 'Remove unused columns'


class ProjectingJoinToProjectOfJoin(Rule):
    """Turn ProjectingJoin to Project of a Join.
    This is useful to take advantage of the column selection
    optimizations and then remove ProjectingJoin for
    backends that don't have one"""

    def fire(self, expr):
        if isinstance(expr, algebra.ProjectingJoin):
            return algebra.Apply([(None, x) for x in expr.output_columns],
                                 algebra.Join(expr.condition,
                                              expr.left,
                                              expr.right))

        return expr

    def __str__(self):
        return 'ProjectingJoin[$1] => Project[$1](Join)'


class RemoveNoOpApply(Rule):
    """Remove Apply operators that have no effect."""

    def fire(self, op):
        if not isinstance(op, algebra.Apply):
            return op

        # At least one emit expression is not just copying a column
        if not all(isinstance(e[1], expression.AttributeRef)
                   for e in op.emitters):
            return op

        child = op.input
        child_scheme = child.scheme()

        # Schemes are different, this Apply does something
        if child_scheme != op.scheme():
            return op

        emitters = [expression.toUnnamed(e[1], child_scheme)
                    for e in op.emitters]
        # Schemes are the same (including names), and this Apply keeps all
        # columns in the same order. This Apply does nothing.
        if all(e.position == i for (i, e) in enumerate(emitters)):
            return child

        return op

    def __str__(self):
        return 'Remove no-op apply'


class SwapJoinSides(Rule):
    # swaps the inputs to a join
    def fire(self, expr):
        # don't allow swap-created join to be swapped
        if isinstance(expr, algebra.Join) and not hasattr(expr, '__swapped__'):
            assert type(expr) is algebra.Join
            # An apply will undo the effect of the swap on the scheme,
            # so above operators won't be affected
            left_sch = expr.left.scheme()
            right_sch = expr.right.scheme()
            leftlen = len(left_sch)
            rightlen = len(right_sch)
            assert leftlen + rightlen == len(expr.scheme())
            emitters_left = [(left_sch.getName(i),
                              UnnamedAttributeRef(rightlen + i))
                             for i in range(leftlen)]
            emitters_right = [(right_sch.getName(i), UnnamedAttributeRef(i))
                              for i in range(rightlen)]
            emitters = emitters_left + emitters_right

            # reindex the expression
            index_map = dict([(oldpos, attr[1].position)
                              for (oldpos, attr) in enumerate(emitters)])

            expression.reindex_expr(expr.condition, index_map)

            newjoin = algebra.Join(expr.condition, expr.right, expr.left)
            newjoin.__swapped__ = True

            return algebra.Apply(emitters=emitters, input=newjoin)
        else:
            return expr

    def __str__(self):
        return "Join(L,R) => Join(R,L)"


# logical groups of catalog transparent rules
# 1. this must be applied first
remove_trivial_sequences = [RemoveTrivialSequences()]

# 2. simple group by
simple_group_by = [SimpleGroupBy()]

# 3. push down selection
push_select = [
    SplitSelects(),
    PushSelects(),
    MergeSelects()
]

# 4. push projection
push_project = [
    ProjectToDistinctColumnSelect(),
    JoinToProjectingJoin()
]

# 5. push apply
push_apply = [
    # These really ought to be run until convergence.
    # For now, run twice and finish with PushApply.
    PushApply(),
    RemoveUnusedColumns(),
    PushApply(),
    RemoveUnusedColumns(),
    PushApply(),
    RemoveNoOpApply(),
]
